import os
import importlib
import random
import string
import copy
import json
import logging
from pprint import pprint, pformat
import troposphere.ssm as ssm

from troposphere import (
    And,
    AWSHelperFn,
    AWSObject,
    AWSProperty,
    Base64,
    BaseAWSObject,
    Condition,
    Equals,
    Export,
    FindInMap,
    GetAtt,
    GetAZs,
    If,
    ImportValue,
    Join,
    Not,
    Or,
    Output,
    Parameter,
    Ref,
    Select,
    Split,
    Sub,
    Tags,
    Template,
)

from . import cfg

from .cfg import (
    MAX_SECURITY_GROUPS,
    SECURITY_GROUPS_DEFAULT,
    PARAMETERS_SKIP_OVERRIDE_CONDITION,
)


# Temporary fix for https://github.com/cloudtools/troposphere/issues/1474
def my_one_of(class_name, properties, property, conditionals):
    if (properties.get(property) not in conditionals and
            not isinstance(properties.get(property), If)):
        raise ValueError(
            # Ensure we handle None as a valid value
            '%s.%s must be one of: "%s"' % (
                class_name, property, ', '.join(
                    condition for condition in conditionals if condition
                )
            )
        )



import troposphere.rds as rds

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)
from .route53 import R53_RecordSetRDS


class RDSDBInstance(rds.DBInstance):
    def validate(self):
        if 'DBSnapshotIdentifier' not in self.properties:
            if 'Engine' not in self.properties:
                raise ValueError(
                    'Resource Engine is required in type %s'
                    % self.resource_type)

#       if 'SourceDBInstanceIdentifier' in self.properties:
#
#           invalid_replica_properties = (
#               'BackupRetentionPeriod', 'DBName', 'MasterUsername',
#               'MasterUserPassword', 'PreferredBackupWindow', 'MultiAZ',
#               'DBSnapshotIdentifier',
#           )
#
#           invalid_properties = [s for s in self.properties.keys() if
#                                 s in invalid_replica_properties]
#
#           if invalid_properties:
#                raise ValueError(
#                   ('{0} properties can\'t be provided when '
#                    'SourceDBInstanceIdentifier is present '
#                    'AWS::RDS::DBInstance.'
#                    ).format(', '.join(sorted(invalid_properties))))

        if ('DBSnapshotIdentifier' not in self.properties and
            'SourceDBInstanceIdentifier' not in self.properties) and \
            ('MasterUsername' not in self.properties or
            'MasterUserPassword' not in self.properties) and \
                ('DBClusterIdentifier' not in self.properties):
            raise ValueError(
                'Either (MasterUsername and MasterUserPassword) or'
                ' DBSnapshotIdentifier are required in type '
                'AWS::RDS::DBInstance.'
            )

        if 'KmsKeyId' in self.properties and \
           'StorageEncrypted' not in self.properties:
            raise ValueError(
                'If KmsKeyId is provided, StorageEncrypted is required '
                'AWS::RDS::DBInstance.'
            )

        nonetype = type(None)
        avail_zone = self.properties.get('AvailabilityZone', None)
        multi_az = self.properties.get('MultiAZ', None)
        if not (isinstance(avail_zone, (AWSHelperFn, nonetype)) and
                isinstance(multi_az, (AWSHelperFn, nonetype))):
            if avail_zone and multi_az in [True, 1, '1', 'true', 'True']:
                raise ValueError("AvailabiltyZone cannot be set on "
                                 "DBInstance if MultiAZ is set to true.")

        storage_type = self.properties.get('StorageType', None)
        if storage_type and storage_type == 'io1' and \
                'Iops' not in self.properties:
            raise ValueError("Must specify Iops if using StorageType io1")

        allocated_storage = self.properties.get('AllocatedStorage')
        iops = self.properties.get('Iops', None)
        if iops and not isinstance(iops, AWSHelperFn):
            if not isinstance(allocated_storage, AWSHelperFn) and \
                    allocated_storage < 100:
                raise ValueError("AllocatedStorage must be at least 100 when "
                                 "Iops is set.")
            if not isinstance(allocated_storage, AWSHelperFn) and not \
                    isinstance(iops, AWSHelperFn) and \
                    float(iops) / float(allocated_storage) > 10.0:
                raise ValueError("AllocatedStorage must be no less than "
                                 "1/10th the provisioned Iops")

        return True


    def setup(self):
        self.AllocatedStorage = get_endvalue('AllocatedStorage')
        self.AllowMajorVersionUpgrade = 'True'
        self.DBInstanceClass = get_endvalue('DBInstanceClass')
        self.DBName = get_endvalue('DBName', nocondition='DBInstanceSkipProperties')
        self.Engine = get_endvalue('Engine')
        self.EngineVersion = get_endvalue('EngineVersion')
        self.MasterUsername = get_endvalue('MasterUsername', nocondition='DBInstanceSkipProperties')
        self.MasterUserPassword = get_endvalue('MasterUserPassword', nocondition='DBInstanceSkipProperties')
        self.MultiAZ = get_endvalue('MultiAZ')
        self.DBParameterGroupName = Ref('DBParameterGroup1')
        self.SourceDBInstanceIdentifier = get_endvalue('SourceDBInstanceIdentifier', condition=True)
        self.StorageType = get_endvalue('StorageType')
        self.VPCSecurityGroups = [Ref('SecurityGroupRDS')]


class RDSDBInstancePublic(RDSDBInstance):
    def setup(self):
        super(RDSDBInstancePublic, self).setup()
        self.DBSubnetGroupName = get_expvalue('DBSubnetGroupPublic')
        self.PubliclyAccessible = 'True'


class RDSDBInstancePrivate(RDSDBInstance):
    def setup(self):
        super(RDSDBInstancePrivate, self).setup()
        self.DBSubnetGroupName = get_expvalue('DBSubnetGroupPrivate')
        self.PubliclyAccessible = 'False'


class RDSDBParameterGroup(rds.DBParameterGroup):
    def setup(self):
        self.Description = Sub('MYSQL %s - ${AWS::StackName}' % self.title)
        self.Family = get_subvalue('mysql${1M}', 'EngineVersion')


class RDSDBSubnetGroupPrivate(rds.DBSubnetGroup):
    def setup(self):
        self.DBSubnetGroupDescription = Sub('${EnvShort}-Private')
        self.SubnetIds=Split(',', get_expvalue('SubnetsPrivate'))


class RDSDBSubnetGroupPublic(rds.DBSubnetGroup):
    def setup(self):
        self.DBSubnetGroupDescription = Sub('${EnvShort}-Public')
        self.SubnetIds=Split(',', get_expvalue('SubnetsPublic'))

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class RDS_ParameterGroups(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = '%s%s' % (key, n)
            # Resources
            r_PG = RDSDBParameterGroup(resname)
            r_PG.setup()
            r_PG.Parameters = v

            add_obj(r_PG)


class RDS_DB(object):
    def __init__(self, key):
        # Resources
        if cfg.RDSScheme == 'External':
            R_DB = RDSDBInstancePublic('DBInstance')
        if cfg.RDSScheme == 'Internal':
            R_DB = RDSDBInstancePrivate('DBInstance')

        R_DB.setup()

        R53_RecordSetRDS()

        add_obj([
            R_DB,
        ])


class RDS_SubnetGroups(object):
    def __init__(self, key):
        # Resources
        R_Private = RDSDBSubnetGroupPrivate('DBSubnetGroupPrivate')
        R_Private.setup()

        R_Public = RDSDBSubnetGroupPublic('DBSubnetGroupPublic')
        R_Public.setup()

        add_obj([
            R_Private,
            R_Public,
        ])

        # Outputs
        O_Private = Output('DBSubnetGroupPrivate')
        O_Private.Value = Ref('DBSubnetGroupPrivate')
        O_Private.Export = Export('DBSubnetGroupPrivate')

        O_Public = Output('DBSubnetGroupPublic')
        O_Public.Value = Ref('DBSubnetGroupPublic')
        O_Public.Export = Export('DBSubnetGroupPublic')

        add_obj([
            O_Private,
            O_Public,
        ])



import troposphere.dynamodb as ddb

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class DDBTableCredStash(ddb.Table):
    def setup(self):
        self.AttributeDefinitions = [
            ddb.AttributeDefinition(
                AttributeName='name',
                AttributeType='S'
            ),
            ddb.AttributeDefinition(
                AttributeName='version',
                AttributeType='S'
            )
        ]
        self.KeySchema = [
            ddb.KeySchema(
                AttributeName='name',
                KeyType='HASH'
            ),
            ddb.KeySchema(
                AttributeName='version',
                KeyType='RANGE'
            )
        ]
        self.ProvisionedThroughput = ddb.ProvisionedThroughput(
            ReadCapacityUnits=1,
            WriteCapacityUnits=1
        )
        self.TableName = Sub('credential-store-${EnvShort}')



import troposphere.servicediscovery  as srvd

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class ServiceDiscoveryPublicDnsNamespace(srvd.PublicDnsNamespace):
    def setup(self):
        self.Description = 'Service Discovery'
        self.Name = Sub('find.' + cfg.HostedZoneNameEnv)


class SRVD_ServiceDiscoveryRES(object):
    def __init__(self, key):
        # Resources
        R_PublicDnsNamespace = ServiceDiscoveryPublicDnsNamespace('PublicDnsNamespace')
        R_PublicDnsNamespace.setup()

        add_obj([
            R_PublicDnsNamespace,
        ])

        # Outputs
        O_PublicDnsNamespace = Output('ServiceDiscoveryPublicDnsNamespaceId')
        O_PublicDnsNamespace.Value = Ref('PublicDnsNamespace')
        O_PublicDnsNamespace.Export = Export('ServiceDiscoveryPublicDnsNamespaceId')

        add_obj([
            O_PublicDnsNamespace,
        ])



import troposphere.logs as lgs

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class LogsLogGroup(lgs.LogGroup):
    def __init__(self, title, **kwargs):
        super(LogsLogGroup, self).__init__(title, **kwargs)
        self.LogGroupName = get_endvalue('LogGroupName')
        self.RetentionInDays = get_endvalue('LogRetentionInDays')


# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class LGS_LogGroup(object):
    def __init__(self, key):
        R_Group = LogsLogGroup('LogsLogGroup')

        add_obj([
            R_Group,
        ])



import troposphere.ssm as ssm

from .common import *


# S - PARAMETERS #
class Parameter(Parameter):
    def __init__(self, title, **kwargs):
        super(Parameter, self).__init__(title, **kwargs)
        self.Type = 'String'
        self.Default = ''


class SSMParameter(ssm.Parameter):
    Type = 'String'
# E - PARAMETERS #


def stack_add_res():
    for n, v in cfg.Parameters.items():
        # Automatically create override conditions for parameters
        if not n.startswith(PARAMETERS_SKIP_OVERRIDE_CONDITION):

            if n == 'InstanceType':
                default = 'default'
            elif n == 'SecurityGroups':
                default = cfg.SECURITY_GROUPS_DEFAULT
            else:
                default = ''

            condition = {
                n + 'Override': Not(Equals(Select(0, Ref(n)), default))
            } if v.Type == 'CommaDelimitedList' else {
                n + 'Override': Not(Equals(Ref(n), default))
            }

            add_obj(condition)
        # End
        cfg.template.add_parameter(v)

    cfg.Parameters.clear()

    for n, v in cfg.Conditions.items():
        cfg.template.add_condition(n, v)
    cfg.Conditions.clear()

    for n, v in cfg.Mappings.items():
        cfg.template.add_mapping(n, v)
    cfg.Mappings.clear()

    for n, v in cfg.Resources.items():
        cfg.template.add_resource(v)
    cfg.Resources.clear()

    for n, v in cfg.Outputs.items():
        cfg.template.add_output(v)
    cfg.Outputs.clear()


def add_obj(obj):
    if isinstance(obj, list):
        for n in obj:
            add_obj(n)
    elif isinstance(obj, Parameter):
        cfg.Parameters[obj.title] = obj
    elif isinstance(obj, dict):
        cfg.Conditions.update(obj)
    elif isinstance(obj, Output):
        cfg.Outputs[obj.title] = obj
    else:
        cfg.Resources[obj.title] = obj


def do_no_override(action):
    if action is True:
        cfg.no_override = True
    else:
        cfg.no_override = False


# def import_modules(gbl):
#     for module in cfg.IMPORT_MODULES:
#         mod = importlib.import_module('.%s' % module, package='awsibox')
#         gbl.update(mod.__dict__)


def _get_value(param, fixedvalues, strout, nolist, issub):
    # Its not pythonic, but it's only way to avoid circular import problems
    from .securitygroup import SG_SecurityGroupsTSK

    # set default if not defined
    if not fixedvalues:
        fixedvalues = cfg.fixedvalues

    # if param in fixedvalues means its value do not changes based on Env/Region so hardcode the value in json, ...
    if param in fixedvalues:
        value = fixedvalues[param]
        # check if value start with method and use eval to run code ... in list too
        if isinstance(value, list):
            value = [eval(r) if r.startswith(cfg.EVAL_FUNCTIONS_IN_CFG) else r for r in value]
        if isinstance(value, str):
            value = eval(value.replace('\n', '')) if value.startswith(cfg.EVAL_FUNCTIONS_IN_CFG) else value
    # ... otherway use mapping
    else:
        value = FindInMap(Ref('EnvShort'), Ref("AWS::Region"), param)

    if strout is True and isinstance(value, int):
        value = str(value)

    if nolist is True and isinstance(value, list):
        value = ','.join(value)

    if issub:
        value = Sub(value)

    return value


def _get_overridevalue(param, value, condname=None):
    if param not in cfg.Parameters and condname in cfg.Parameters:
        param = condname

    if (cfg.no_override is False and param in cfg.Parameters and not
            param.startswith(PARAMETERS_SKIP_OVERRIDE_CONDITION)):
        override_value = If(param + 'Override', Ref(param), value)
    else:
        override_value = value

    return override_value


def get_endvalue(param, ssm=False, condition=False, nocondition=False,
    nolist=False, inlist=False, split=False, issub=False, strout=False,
    fixedvalues=None, mapinlist=False,
):

    value = _get_value(param, fixedvalues, strout, nolist, issub)

    if mapinlist is not False:
        value = Select(mapinlist, value) 

    if condition or nocondition:
        if condition is True or nocondition is True:
            condname = param
        elif condition:
            condname = condition
        elif nocondition:
            condname = nocondition

        value = _get_overridevalue(param, value, condname)

        value = If(
            condname,
            value if condition else Ref('AWS::NoValue'),
            Ref('AWS::NoValue') if condition else value,
        )
    else:
        value = _get_overridevalue(param, value)

    if split is not False:
        value = Select(split, Split(',', value))

    if inlist is not False:
        value = Select(inlist, value)

    return value


def get_expvalue(param, stack=False, prefix=''):
    v = ''
    if stack:
        v = ImportValue(
            Sub('%s-${%s}' % (param, stack), **{stack: get_endvalue(stack)})
        )
    elif not isinstance(param, str):
        v = ImportValue(
            Sub(prefix + '${ImportName}', **{'ImportName': param})
        )
    else:
        v = ImportValue(param)

    return v


def get_subvalue(substring, subvar, stack=False):
    submap = {}
    found = substring.find('${')

    while found != -1:
        posindex = found + 2
        myindex = substring[posindex]
        mytype = substring[posindex + 1]
        # M = Mapped, E = Exported
        if myindex.isdigit() and mytype in ['M', 'E']:
            listitem = subvar[int(myindex) - 1] if isinstance(subvar, list) else subvar
            stackitem = stack[int(myindex) - 1] if isinstance(stack, list) else stack
            if mytype == 'M':
                submap[listitem] = get_endvalue(listitem)
            else:
                submap[listitem] = get_expvalue(listitem, stackitem)
            substring = substring.replace('${' + myindex + mytype + '}', '${' + listitem + '}')
        found = substring.find('${', posindex)

    v = Sub(substring, **submap)

    return v


def get_condition(cond_name, cond, value2, key=None, OrExtend=[], mapinlist=False):
    # record current state
    override_state = cfg.no_override
    do_no_override(True)

    key_name = key if key else cond_name
    if isinstance(key, FindInMap):
        map_name = key.data['Fn::FindInMap'][0]
        key_name = key.data['Fn::FindInMap'][1] 
        value_name = key.data['Fn::FindInMap'][2]
        if not value_name and cond_name:
            value_name = cond_name

        value1_param = FindInMap(map_name, Ref(key_name), value_name)
        value1_map = FindInMap(map_name, get_endvalue(key_name), value_name)
    elif isinstance(key, Select):
        select_index = key.data['Fn::Select'][0]
        select_list = key.data['Fn::Select'][1]
        split_sep = select_list.data['Fn::Split'][0]
        key_name = select_list.data['Fn::Split'][1]

        value1_param = Select(select_index, Split(split_sep, Ref(key_name)))
        value1_map = Select(select_index, Split(split_sep, get_endvalue(key_name)))
    else:
        value1_param = Ref(key_name)
        # Used new param "mapinlist" when you have a mapped value in a list but multiple values as override parameters
        if mapinlist:
            value1_map = get_endvalue(mapinlist[0], mapinlist=mapinlist[1])
        else:
            value1_map = get_endvalue(key_name)

    # if beginning state was False set it back
    if not override_state:
        do_no_override(False)

    eq_param = Equals(value1_param, value2)
    eq_map = Equals(value1_map, value2)

    if cond == 'equals':
        cond_param = eq_param
        cond_map = eq_map
    elif cond == 'not_equals':
        cond_param = Not(eq_param)
        cond_map = Not(eq_map)

    if (key_name in cfg.Parameters and not
            key_name.startswith(PARAMETERS_SKIP_OVERRIDE_CONDITION)):
        key_override = key_name + 'Override'
        condition = Or(
            And(
                Condition(key_override),
                cond_param,
            ),
            And(
                Not(Condition(key_override)),
                cond_map,
            )
        )
        if OrExtend:
            condition.data['Fn::Or'].extend(OrExtend)
    else:
        condition = cond_map

    if cond_name:
        condition = {cond_name: condition}

    return condition


def import_lambda(name):
    parent_dir_name = os.path.dirname(os.path.realpath(__file__))
    lambda_file = os.path.join(os.getcwd(), 'lib/lambdas/%s.code' % name)
    if not os.path.exists(lambda_file):
        lambda_file = os.path.join(parent_dir_name, 'lambdas/%s.code' % name)
    try:
        with open(lambda_file, 'r') as f:
            file_lines = []
            for x in f.readlines():
                if x.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
                    value = eval(x)
                elif x.startswith('STRING_EVAL_FUNCTIONS_IN_CFG'):
                    value = '"'
                else:
                    value = ''.join(x)

                file_lines.append(value)

            return(file_lines)

    except IOError:
        logging.error('Lambda code %s not found' % name)
        exit(1)


def auto_get_props_recurse(obj, key, props, obj_propname, mapname, propname, rootkey=None, rootname=None):
    prop_class = props[obj_propname][0]
    if isinstance(prop_class, type) and prop_class.__bases__[0].__name__ == 'AWSProperty':
        # If object already have that props, object class is the existing already defined object, so extend its property with auto_get_props
        if obj_propname in obj.properties:
            prop_obj = obj.properties[obj_propname]
        else:
            prop_obj = prop_class()
        auto_get_props(
            prop_obj,
            key=key[obj_propname],
            mapname=mapname + obj_propname,
            recurse=True,
            rootkey=rootkey[obj_propname] if rootkey else None,
            rootname=rootname + obj_propname if rootname else None,
        )

        return prop_obj

    elif isinstance(prop_class, list) and isinstance(prop_class[0], type) and prop_class[0].__bases__[0].__name__ == 'AWSProperty':
        # If object props already is a list, keep existing list objects
        # need to make a change, if object already exist, i must first iterate over old list with rootkey and rootname and then over new one
        prop_list = []
        prop_class = props[obj_propname][0][0]
        # If rootkey is defined, first iterate over rootkey and execute auto_get_props passing rootkey and rootname, but check if element exist in key too,
        # in that case execute execute auto_get_props passing key and mapname...
        if rootkey:
            for o, v in rootkey[obj_propname].items():
                name_o = str(o)
                rootname_o = rootname + obj_propname + name_o
                mapname_o = mapname + obj_propname + name_o
                prop_obj = prop_class()
                auto_get_props(
                    prop_obj,
                    key=v,
                    mapname=rootname_o,
                    recurse=True,
                    rootkey=v,
                    rootname=rootname_o,
                )
                if o in key[obj_propname]:
                    key_o = key[obj_propname][o]
                    auto_get_props(
                        prop_obj,
                        key=key_o,
                        mapname=mapname_o,
                        recurse=True,
                        rootkey=key_o,
                        rootname=rootname_o,
                    )
                prop_list.append(prop_obj)
        # ...then iterate over key, but check if element exist in rootkey too, in that case skip it (already included in previous rootkey iteration)
        # when calling auto_get_props rootkey and rootname can be setted to None, cause if element has not been skipped mean that it do not exist in rootkey
        # so there is no need to pass it (there cannot be a corrispective node in rootkey)
        for o, v in key[obj_propname].items():
            name_o = str(o)
            mapname_o = mapname + obj_propname + name_o
            prop_obj = prop_class()
            if rootkey and o in rootkey[obj_propname]:
                continue
            auto_get_props(
                prop_obj,
                key=v,
                mapname=mapname_o,
                recurse=True,
                rootkey=None,
                rootname=None,
            )
            prop_list.append(prop_obj)

        return prop_list

    return get_endvalue(mapname + propname)


def auto_get_props(obj, key=None, del_prefix='', mapname=None, recurse=False, rootkey=None, rootname=None, rootdict=None):
    # set default if not defined
    if not key:
        key = cfg.RP_cmm

    # Allowed object properties
    props = vars(obj)['propnames']
    props.extend(vars(obj)['attributes'])
    # Object class
    classname = obj.__class__.__name__

    # build up mapname
    mapname = mapname if mapname is not None else obj.title
    if classname in ['Output', 'Parameter']:
        mapname = classname + mapname

    # iterate over props or key choosing the one with less objects
    use_key = True if len(props) > len(key) else None
    for propname in key if use_key else props:
        obj_propname = propname.replace(del_prefix, '') if use_key else del_prefix + propname
        if obj_propname in (props if use_key else key):
            if recurse and isinstance(key[obj_propname], dict):
                value = auto_get_props_recurse(obj, key, obj.props, obj_propname, mapname, propname, rootkey, rootname)
            # needed for lib/efs.py EFS_FileStorage SGIExtra - where is passed as key a new dictionary to parse for parameters
            elif rootdict:
                value = get_endvalue(mapname + propname, fixedvalues=rootdict)
            else:
                value = get_endvalue(mapname + propname)

            # Avoid intercepting a Template Condition as a Resource Condition
            if obj_propname == 'Condition' and not isinstance(value, str):
                continue

            try:
                setattr(obj, obj_propname, value)
            except TypeError:
                pass


def auto_build_obj(obj, key):
    props = vars(obj)['propnames']
    classname = obj.__class__
    for resname, resvalue in key.items():
        final_obj = classname(resname)
        auto_get_props(final_obj, resvalue)

        add_obj(final_obj)


def change_obj_data(obj, find, value):
    if isinstance(obj, list):
        for n, v in enumerate(obj):
            change_obj_data(obj[n], find, value)
    elif isinstance(obj, If):
        obj_if = obj.data['Fn::If']
        for n, v in enumerate(obj_if):
            if isinstance(v, Ref) and v.data['Ref'] == find:
                obj_if[n] = value


def gen_random_string():
    length = 16
    char_set = string.ascii_letters + string.digits
    if not hasattr(gen_random_string, "rng"):
        gen_random_string.rng = random.SystemRandom()  # Create a static variable
    secret_string = ''.join([gen_random_string.rng.choice(char_set) for _ in range(length)])

    return secret_string



import troposphere.cloudwatch as clw

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


class CW_Alarms(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            if not ('Enabled' in v and v['Enabled'] is True):
                continue
            resname = '%s%s' % (key, n)
            # parameters
            p_EvaluationPeriods = Parameter(resname + 'EvaluationPeriods')
            p_EvaluationPeriods.Description = 'Number of periods for alarm evaluation - 0 to disable - empty for default based on env/role'
            p_EvaluationPeriods.AllowedValues = ['', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

            p_Period = Parameter(resname + 'Period')
            p_Period.Description = 'Period lenght in seconds (multiple of 60) - empty for default based on env/role'
            p_Period.AllowedValues = ['', '60', '120', '180', '240', '300']

            p_Threshold = Parameter(resname + 'Threshold')
            p_Threshold.Description = 'Threshold for alarm triggering - empty for default based on env/role'
            p_Threshold.AllowedValues = ['', '10', '15', '20', '25', '30', '35', '40', '45', '50', '55', '60', '65', '70', '75', '80', '85', '90', '95', '100']

            add_obj([
                p_EvaluationPeriods,
                #p_Period,
                p_Threshold,
            ])

            # conditions
            add_obj(get_condition(resname, 'not_equals', '0', resname + 'EvaluationPeriods'))

            # resources
            r_Alarm = clw.Alarm(resname)
            auto_get_props(r_Alarm, v, recurse=True)
            if hasattr(r_Alarm, 'Metrics'):
                r_Alarm.Period = Ref('AWS::NoValue')
            if not hasattr(r_Alarm, 'Condition'):
                r_Alarm.Condition = resname

            add_obj(r_Alarm)

            # outputs
            o_Alarm = Output(resname)
            o_Alarm.Value = get_subvalue(
                'Period=${1M},EvaluationPeriods=${2M},Threshold=${3M}',
                [
                    resname + 'Period',
                    resname + 'EvaluationPeriods',
                    resname + 'Threshold'
                ]
            )

            add_obj(o_Alarm)



import sys
import os
import copy
from collections import OrderedDict, Mapping
from pprint import pprint, pformat

from . import cfg


def mappings_create_entry(mappings, keyenv, keyregion, keyname, keyvalue):
    global mappedvalues

    # create mapping entry
    mappings[keyenv][keyregion].update({keyname: keyvalue})
    if keyenv != 'cmm':
        mappedvalues.add(keyname)
    # and if present delete the common one, but first record its value
    if keyenv != 'cmm' and keyname in mappings['cmm']['cmm']:
        value = mappings['cmm']['cmm'][keyname]
        del mappings['cmm']['cmm'][keyname]
        # for all real env region that do not already have a key entry i need to create a mapping entry equals to the common one
        # for the non already processed one it could be rewritten later.
        for env in cfg.RP_base:
            for region in cfg.RP_base[env]:
                if env != 'cmm' and keyname not in mappings[env][region]:
                    mappings[env][region].update({keyname: value})


def get_mapping_env_region(MP, RP, e, r, p):
    mappings = MP

    if isinstance(RP, dict):
        for i, v in RP.items():
            if not e and not r:
                get_mapping_env_region(mappings, v, i, None, None)
            elif not r:
                get_mapping_env_region(mappings, v, e, i, None)
            elif not p:
                get_mapping_env_region(mappings, v, e, r, i)
            else:
                get_mapping_env_region(mappings, v, e, r, '%s%s' % (p, i))
    else:
        if p in mappings['cmm']['cmm'] and RP == mappings['cmm']['cmm'][p]:
            pass
        else:
            mappings_create_entry(mappings, e, r, p, RP)

    return mappings


def get_envregion_mapping():
    global mappedvalues
    mappedvalues = set()

    mappings = get_mapping_env_region(copy.deepcopy(cfg.RP_base), cfg.RP, None, None, None)
    fixedvalues = mappings['cmm']['cmm']
    cfg.fixedvalues = fixedvalues
    cfg.mappedvalues = mappedvalues

    if cfg.debug:
        print('##########FIXEDVALUES#########START#####')
        pprint(cfg.fixedvalues)
        print('##########FIXEDVALUES#########END#######')

        print('##########MAPPEDVALUES#########START#####')
        pprint(cfg.mappedvalues)
        print('##########MAPPEDVALUES#########END#######')

    # delete empy mappings, CloudFormation do not like them!
    for env in cfg.RP_base:
        for region in cfg.RP_base[env]:
            if len(mappings[env][region]) == 0:
                del mappings[env][region]
        if len(mappings[env]) == 0:
            del mappings[env]

    # and remove no more needed cmm one
    if 'cmm' in mappings:
        del mappings['cmm']

    return mappings


def get_ec2_mapping():
    mappings = {}

#   #START##MAP INSTANCE TYPE EPHEMERAL####
    map_eph = {
        'InstaceEphemeral0': ['m3.medium', 'm3.large', 'm3.xlarge', 'm3.2xlarge', 'c3.large',
                              'c3.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'g2.2xlarge', 'r3.large', 'r3.xlarge',
                              'r3.2xlarge', 'r3.4xlarge', 'i2.xlarge', 'i2.2xlarge', 'i2.4xlarge',
                              'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge'],
        'InstaceEphemeral1': ['m3.xlarge', 'm3.2xlarge', 'c3.large', 'c3.xlarge', 'c3.2xlarge', 'c3.4xlarge',
                              'i2.2xlarge', 'i2.4xlarge', 'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge'],
        'InstaceEphemeral2': ['i2.4xlarge', 'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge']
    }

    mappings['InstanceTypes'] = {}

    for i in cfg.INSTANCE_LIST:
        mappings['InstanceTypes'].update({
            i: {
                'InstaceEphemeral0': '1' if i in map_eph['InstaceEphemeral0'] else '0',
                'InstaceEphemeral1': '1' if i in map_eph['InstaceEphemeral1'] else '0',
                'InstaceEphemeral2': '1' if i in map_eph['InstaceEphemeral2'] else '0',
            }
        })
#   #END##MAP INSTANCE TYPE EPHEMERAL####

    return mappings


def get_azones_mapping():
    mappings = {}
    mappings['AvabilityZones'] = {}
    AZ = cfg.AZones

    for r in cfg.regions:
        zones = {}
        try:
            nzones = AZ[r]
        except:
            nzones = AZ['default']

        try:
            nzones = cfg.RP['dev'][r]['AZones']
        except:
            pass

        for n in range(AZ['MAX']):
            zones['Zone%s' % n] = 'True' if nzones > n else 'False'

        mappings['AvabilityZones'][r] = zones

    return mappings


class Mappings(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            if n == 'EnvRegion':
                mapping = get_envregion_mapping()
            if n == 'EC2':
                mapping = get_ec2_mapping()
            if n == 'AZones':
                mapping = get_azones_mapping()

            cfg.Mappings.update(mapping)



#!/usr/bin/python
from . import cfg
from .shared import stack_add_res
from . import (mappings, lambdas, securitygroup, cloudwatch, loadbalancing, autoscaling, iam,
    codedeploy, route53, crm, events, cloudfront, sqs, sns,
    ecs, ecr, s3, waf, vpc, dynamodb, kms, rds, efs, elasticache,
    servicediscovery, cloudformation, logs, apigateway)


def execute_class(RP_cmm):
    for k, v in cfg.CFG_TO_CLASS.items():
        class_name = v['class']
        module_name = v['module']
        module = globals()[module_name]

        if k in list(RP_cmm.keys()):
            RP_value = RP_cmm[k]
            if isinstance(RP_value, str) and RP_value == 'SkipClass':
                continue
            if isinstance(class_name, list):
                for n in class_name:
                    getattr(module, n)(key=k)
                continue
            stacktype_class = class_name + cfg.stacktype.upper()
            if stacktype_class in dir(module):
                getattr(module, stacktype_class)(key=k)
            elif class_name in dir(module):
                getattr(module, class_name)(key=k)

    stack_add_res()


def generate():
    classenvrole = cfg.envrole.replace('-', '_')  # Ex client-portal -> client_portal
    cfg.classenvrole = classenvrole

    execute_class(cfg.RP_cmm)

    cfg.template.add_description('%s [%s]' % (cfg.envrole, cfg.stacktype))
    cfg.template.add_version('2010-09-09')

    return cfg.template



import troposphere.ecr as ecr

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)
from .securitygroup import (SecurityGroupEcsService, SecurityGroupRuleEcsService,
    SG_SecurityGroupsECS)


class ECRRepositories(ecr.Repository):
    def setup(self):
        self.RepositoryName = get_endvalue(self.title)
        self.RepositoryPolicyText = {
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Action': [
                        'ecr:GetDownloadUrlForLayer',
                        'ecr:BatchGetImage',
                        'ecr:BatchCheckLayerAvailability',
                        'ecr:PutImage',
                        'ecr:InitiateLayerUpload',
                        'ecr:UploadLayerPart',
                        'ecr:CompleteLayerUpload',
                        'ecr:ListImages',
                        'ecr:DescribeRepositories',
                        'ecr:DescribeImages'
                    ],
                    'Effect': 'Allow',
                    'Principal': {
                        'AWS': [
                            Sub('arn:aws:iam::${AWS::AccountId}:root')
                        ]
                    },
                    'Sid': 'AllowPushPull'
                },
            ]
        }

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

class ECRRepositoryLifecyclePolicy(ecr.LifecyclePolicy):
    def __init__(self, title, **kwargs):
        super(ECRRepositoryLifecyclePolicy, self).__init__(title, **kwargs)
        LifecyclePolicyText = {
            'rules': [
                {
                    'action': {
                        'type': 'expire'
                    },
                    'rulePriority': 1,
                    'selection': {
                        'countNumber': 9500,
                        'countType': 'imageCountMoreThan',
                        'tagStatus': 'any',
                    },
                    'description': 'Images are sorted on pushed_at_time (desc), images greater than specified count are expired.'
                }
            ]
        }
        self.LifecyclePolicyText = json.dumps(LifecyclePolicyText)
        self.RegistryId = Ref('AWS::AccountId')


def ECRRepositoryPolicyStatementAccountPull(name):
    policy = {   
        'Action': [
            'ecr:GetDownloadUrlForLayer',
            'ecr:BatchGetImage',
            'ecr:BatchCheckLayerAvailability',
            'ecr:ListImages',
            'ecr:DescribeRepositories',
            'ecr:DescribeImages'
        ],
        'Effect': 'Allow',
        'Principal': {
            'AWS': [
                get_subvalue('arn:aws:iam::${1M}:root', name)
            ]
        },
        'Sid': 'AllowPull'
    }

    return policy


def ECRRepositoryPolicyStatementAccountPush(name):
    policy = {   
        'Action': [
            'ecr:PutImage',
            'ecr:InitiateLayerUpload',
            'ecr:UploadLayerPart',
            'ecr:CompleteLayerUpload',
        ],
        'Effect': 'Allow',
        'Principal': {
            'AWS': [
                get_subvalue('arn:aws:iam::${1M}:root', name)
            ]
        },
        'Sid': 'AllowPush'
    }

    return policy


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################

class ECR_Repositories(object):
    def __init__(self, key):
        PolicyStatementAccounts = []
        for n, v in cfg.EcrAccount.items():
            mapname = 'EcrAccount' + n  + 'Id'  # Ex. EcrAccountPrdId
            # conditions
            add_obj(get_condition(mapname, 'not_equals', 'None'))

            if 'Pull' in v['Policy']:
                PolicyStatementAccount = ECRRepositoryPolicyStatementAccountPull(name=mapname)
                PolicyStatementAccounts.append(
                    If(
                        mapname,
                        PolicyStatementAccount,
                        Ref('AWS::NoValue')
                    )
                )

            if 'Push' in v['Policy']:
                PolicyStatementAccount = ECRRepositoryPolicyStatementAccountPush(name=mapname)
                PolicyStatementAccounts.append(
                    If(
                        mapname,
                        PolicyStatementAccount,
                        Ref('AWS::NoValue')
                    )
                )

        # Resources
        for n, v in getattr(cfg, key).items():
            Repo = ECRRepositories(key + n)  # Ex. RepositoryApiLocationHierarchy
            Repo.setup()
            Repo.RepositoryPolicyText['Statement'].extend(PolicyStatementAccounts)
            Repo.LifecyclePolicy = ECRRepositoryLifecyclePolicy('')

            add_obj(Repo)



import troposphere.awslambda as lbd

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj, import_lambda)
from .iam import IAMRoleLambdaBase

class LambdaPermission(lbd.Permission):
    def setup(self):
        self.Action = 'lambda:InvokeFunction'


class LambdaPermissionEvent(LambdaPermission):
    def setup(self, key, source):
        super(LambdaPermissionEvent, self).setup()
        self.Principal = 'events.amazonaws.com'
        self.FunctionName = eval(key['Arn'])
        self.SourceArn = GetAtt(source, 'Arn')


class LambdaPermissionS3(LambdaPermission):
    def setup(self, key, source):
        super(LambdaPermissionS3, self).setup()
        self.Principal = 's3.amazonaws.com'
        self.FunctionName = key
        self.SourceArn = Sub('arn:aws:s3:::%s' % getattr(cfg, source))


class LambdaPermissionSNS(LambdaPermission):
    def setup(self, key, **kwargs):
        super(LambdaPermissionSNS, self).setup(**kwargs)
        auto_get_props(self, key)
        self.Principal = 'sns.amazonaws.com'
        self.FunctionName = eval(key['Endpoint'])
        self.SourceArn = eval(key['TopicArn'])


class LambdaPermissionApiGateway(LambdaPermission):
    def setup(self, name, source):
        super(LambdaPermissionApiGateway, self).setup()
        self.Principal = 'apigateway.amazonaws.com'
        self.FunctionName = name
        self.SourceArn = source


class LambdaPermissionLoadBalancing(LambdaPermission):
    def setup(self, name):
        super(LambdaPermissionLoadBalancing, self).setup()
        self.Principal = 'elasticloadbalancing.amazonaws.com'
        self.FunctionName = name


class LambdaVersion(lbd.Version):
    def __init__(self, title, name, **kwargs):
        super(LambdaVersion, self).__init__(title, **kwargs)
        self.FunctionName = Ref(name)


class LambdaFunction(lbd.Function):
    def setup(self, key, name):
        import_name = key['ImportName'] if 'ImportName' in key else name
        if 'Code' not in key:
            self.Code = lbd.Code(
                ZipFile=Join('', import_lambda(import_name))
            )
        auto_get_props(self, key, recurse=True)
        self.FunctionName = Sub('${AWS::StackName}-${EnvRole}-' + name)
        if 'Handler' not in key:
            self.Handler = 'index.lambda_handler'
        self.Role = GetAtt('RoleLambda' + name, 'Arn')

        if all(k in key for k in ['SecurityGroupIds', 'SubnetIds']):
            self.VpcConfig = lbd.VPCConfig('')
            auto_get_props(self.VpcConfig, key, mapname=self.title)

        if 'Variables' in key:
            self.Environment = lbd.Environment(
                Variables={
                    varname: get_endvalue(self.title + 'Variables' + varname) for varname in key['Variables']
                }
            )


class LambdaLayerVersionPermission(lbd.LayerVersionPermission):
    def __init__(self, title, **kwargs):
        super(LambdaLayerVersionPermission, self).__init__(title, **kwargs)
        self.Action = 'lambda:GetLayerVersion'
        self.Principal = Ref('AWS::AccountId')


def LambdaLayers(obj, resname, i):
    layername = '%s%s' % (resname, i)
    # parameters
    p_Layer = Parameter(layername)
    p_Layer.Description = layername

    add_obj(p_Layer)

    # condition
    add_obj(
        get_condition(layername, 'not_equals', '', mapinlist=(resname, i))
    )

    # output
    o_Layer = Output(layername)
    o_Layer.Value = Select(i, get_endvalue(resname, condition=layername))
    o_Layer.Value = get_endvalue(resname, condition=layername, mapinlist=i)

    return o_Layer.Value

##

class LBD_Lambdas(object):
    def __init__(self, key):
        # Resources
        for n, v in getattr(cfg, key).items():
            resname = key + n

            try: v['Code']['S3Key']
            except: pass
            else:
                s3keyname = resname + 'Code' + 'S3Key'
                # parameters
                p_S3Key = Parameter(s3keyname)
                p_S3Key.Description = 'S3Key Name for lambda %s Code' % n

                add_obj(p_S3Key)

                # outputs
                o_S3Key = Output(s3keyname)
                o_S3Key.Value = get_endvalue(s3keyname)

                add_obj(o_S3Key)

            # resources
            r_Lambda = LambdaFunction(resname)
            r_Lambda.setup(key=v, name=n)

            if 'Enabled' in v:
                # conditions
                add_obj(get_condition(resname, 'not_equals', 'None', resname + 'Enabled'))

                r_Lambda.Condition = resname

            if 'Layers' in v:
                r_Lambda.Layers = []
                for i, j in enumerate(v['Layers']):
                    r_Lambda.Layers.append(
                        LambdaLayers(r_Lambda, '%s%s' % (resname, 'Layers'), i)
                    )

            if 'Version' in v:
                versionname = resname + 'Version'
                # parameters
                p_Version = Parameter(versionname)
                p_Version.Description = 'LambdaVersion change between A/B to force deploy new version'
                p_Version.AllowedValues = ['', 'A', 'B']
                p_Version.Default = ''

                add_obj(p_Version)

                # conditons
                add_obj([
                    {versionname + 'A': Equals(
                        Ref(resname + 'Version'), 'A'
                    )},
                    {versionname + 'B': Equals(
                        Ref(resname + 'Version'), 'B'
                    )},
                    {versionname: Or(
                        Condition(versionname + 'A'),
                        Condition(versionname + 'B'),
                    )},
                ])

                # resources
                r_VersionA = LambdaVersion(versionname + 'A', name=resname )
                r_VersionA.Condition = versionname + 'A'
                
                r_VersionB = LambdaVersion(versionname + 'B', name=resname )
                r_VersionB.Condition = versionname + 'B'

                add_obj([
                    r_VersionA,
                    r_VersionB,
                ])

                # outputs
                o_Version = Output(versionname)
                o_Version.Value = If(
                    versionname + 'A',
                    Ref(versionname + 'A'),
                    Ref(versionname + 'B')
                )
                o_Version.Condition = versionname
                
                add_obj([
                    o_Version,
                ])

            # Automatically setup a lambda Role with base permissions.
            r_Role = IAMRoleLambdaBase('Role' + resname)
            r_Role.setup(key=v)
            if hasattr(r_Lambda, 'Condition'):
            	r_Role.Condition = r_Lambda.Condition

            add_obj([
                r_Lambda,
                r_Role,
            ])

            if 'Export' in v and v['Export']:
                O_Lambda = Output(resname)
                O_Lambda.Value = GetAtt(resname, 'Arn')
                O_Lambda.Export = Export(resname)

                add_obj(O_Lambda)


class LBD_LayerVersions(object):
    def __init__(self, key):
        # Resources
        for n, v in getattr(cfg, key).items():
            resname = key + n

            try: v['Content']['S3Key']
            except: pass
            else:
                s3keyname = resname + 'Content' + 'S3Key'
                # parameters
                p_S3Key = Parameter(s3keyname)
                p_S3Key.Description = 'S3Key Name for lambda %s Content' % n

                add_obj(p_S3Key)

                # outputs
                o_S3Key = Output(s3keyname)
                o_S3Key.Value = get_endvalue(s3keyname)

                add_obj(o_S3Key)

            # resources
            r_Layer = lbd.LayerVersion(resname)
            auto_get_props(r_Layer, v, recurse=True)
            r_LayerPermission = LambdaLayerVersionPermission('LambdaLayerPermission' + n)
            r_LayerPermission.LayerVersionArn = Ref(resname)

            add_obj([
                r_Layer,
                r_LayerPermission,
            ])

            # output
            o_Layer = Output(resname)
            o_Layer.Value = Ref(resname)

            add_obj(o_Layer)


class LBD_Permissions(object):
    def __init__(self, key):
        # Resources
        for n, v in getattr(cfg, key).items():
            resname = key + n

            # resources
            r_Permission = LambdaPermission(resname)
            r_Permission.setup()
            auto_get_props(r_Permission, v, recurse=True)

            add_obj([
                r_Permission,
            ])


class LBD_EventSourceMappings(object):
    def __init__(self, key):
        # Resources
        for n, v in getattr(cfg, key).items():
            resname = key + n

            # resources
            r_EventSourceMapping = lbd.EventSourceMapping(resname)
            auto_get_props(r_EventSourceMapping, v, recurse=True)

            add_obj([
                r_EventSourceMapping,
            ])



import troposphere.iam as iam

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj, SSMParameter)


class IAMUser(iam.User):
    def setup(self, key, name):
        self.Condition = self.title
        self.UserName = key['UserName']
        self.LoginProfile = iam.LoginProfile(
            Password=GetAtt('SSMParameterPassword' + name, 'Value'),
            PasswordResetRequired=True
        )


class IAMGroup(iam.Group):
    def setup(self, key, name):
        self.Condition = self.title
        self.GroupName = name


class IAMUserToGroupAddition(iam.UserToGroupAddition):
    def setup(self, key, name):
        # self.Condition = name
        self.GroupName = name


class IAMPolicy(iam.PolicyType):
    def setup(self, key, name):
        self.PolicyName = name
        auto_get_props(self, key)
        self.PolicyDocument = {
            'Version': '2012-10-17',
        }


class IAMManagedPolicy(iam.ManagedPolicy):
    def setup(self, key, name):
        auto_get_props(self, key)
        self.PolicyDocument = {
            'Version': '2012-10-17',
        }
        self.Description = key['Description']


class IAMPolicyBucketReplica(iam.PolicyType):
    def setup(self, bucket, bucket_name, key):
        name = self.title  # Ex. IAMPolicyReplicaBucketElasticSearch
        self.Condition = bucket + 'Replica'
        self.PolicyName = self.title
        self.PolicyDocument = {
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Action': [
                        's3:GetReplicationConfiguration',
                        's3:ListBucket'
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        Sub('arn:aws:s3:::%s' % bucket_name)
                    ]
                },
                {
                    'Action': [
                        's3:GetObjectVersion',
                        's3:GetObjectVersionAcl',
                        's3:GetObjectVersionTagging',
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        Sub('arn:aws:s3:::%s/*' % bucket_name) 
                    ]
                },
                {
                    'Action': [
                        's3:ReplicateObject',
                        's3:ReplicateDelete',
                        's3:ReplicateTags',
                        's3:ObjectOwnerOverrideToBucketOwner',
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        get_subvalue(
                            'arn:aws:s3:::${1M}/*',
                            bucket + 'ReplicaDstBucket'
                        ) if 'ReplicaDstBucket' in key else get_subvalue(
                            'arn:aws:s3:::${1M}-%s/*' % bucket_name.replace('${AWS::Region}-', '', 1),
                            bucket + 'ReplicaDstRegion',
                        )
                    ]
                }
            ]
        }
        self.Roles = [
            Ref('Role' + self.Condition)  # Ex. RoleBucketImagesReplica
        ]


class IAMInstanceProfile(iam.InstanceProfile):
    def setup(self):
        self.Path = '/'
        self.Roles=[
            Ref('RoleInstance')
        ]


class IAMRoleIBox(iam.Role):
    def setup(self):
        # Trick to add a policy from cfg/yml to a role created from code (Ex. LambdaRoles)
        try:
            cfg.IAMPolicyInRole
        except:
            pass
        else:
            for n, v in cfg.IAMPolicyInRole.items():
                if self.title.endswith(n):
                    self.Policies = [
                        IAMPolicyInRole(n, v)
                    ]
                    break


class IAMRole(iam.Role):
    def setup(self, key):
        auto_get_props(self, key)
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Effect': 'Allow',
                'Principal': {
                    key['PrincipalType'] if 'PrincipalType' in key else 'Service': [get_endvalue(self.title + 'Principal')]
                },
            }]
        }
        self.Path = '/'


class IAMRoleLambdaBase(IAMRoleIBox):
    def setup(self, key):
        super(IAMRoleLambdaBase, self).setup()
        self.Path = '/'
        self.ManagedPolicyArns = [
            'arn:aws:iam::aws:policy/AmazonEC2ReadOnlyAccess',
            'arn:aws:iam::aws:policy/service-role/AWSLambdaVPCAccessExecutionRole'
        ]
        if 'RoleManagedPolicyArns' in key:
            self.ManagedPolicyArns.extend(get_endvalue('RoleManagedPolicyArns', fixedvalues=key))
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Principal': {'Service': [
                    'lambda.amazonaws.com',
                    'edgelambda.amazonaws.com' if 'AtEdge' in key else Ref('AWS::NoValue')
                ]},
                'Effect': 'Allow'
            }]
        }


class IAMRoleBucketReplica(iam.Role):
    def setup(self):
        self.Condition = self.title.replace('Role','')  # Ex. BucketPortalStaticReplica
        self.Path = '/'
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Effect': 'Allow',
                'Principal': {'Service': ['s3.amazonaws.com']},
            }],
            'Version': '2012-10-17'
        }


# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################


def IAMPolicyInRole(name, key):
    Policy = iam.Policy('')
    Policy.PolicyName = name
    Policy.PolicyDocument = {
            'Version': '2012-10-17',
            }
    Statement = []
    for n, v in key['Statement'].items():
        Statement.append(IAMPolicyStatement(v))
    Policy.PolicyDocument['Statement'] = Statement

    return Policy


def IAMPolicyStatement(key):
    Statement = {
        'Effect': key['Effect'] if 'Effect' in key else 'Allow',
    }

    for k in ['Action', 'NotAction', 'Resource', 'NotResource', 'Condition']:
        if k in key:
            Statement.update({k: get_endvalue(k, fixedvalues=key)})

    return Statement

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class IAM_Users(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = key + n  # Ex. IAMUserPincoPalla

            # parameters
            p_Password = Parameter('PasswordBase' + n)
            p_Password.Description = 'Base Password, must be changed at first login'
            p_Password.Default = ''
            p_Password.NoEcho = True

            add_obj(p_Password)

            # conditions
            add_obj([
                get_condition(resname, 'not_equals', 'None', resname + 'Enabled'),
                get_condition(resname + 'RoleAccount', 'not_equals', 'None'),
            ])

            ManagedPolicyArns = []
            RoleGroups = []
            if 'RoleGroups' in v:
                for m, w in v['RoleGroups'].items():
                    condname = resname + 'RoleGroups' + m
                    # conditions
                    add_obj(get_condition(condname, 'not_equals', 'None'))
    
                    # resources
                    RoleGroups.append(
                        If(
                            condname,
                            m,
                            Ref('AWS::NoValue')
                        )
                    )
   
                    try: policy_arns = cfg.IAMGroup[m]['ManagedPolicyArns']
                    except:
                        pass
                    else:
                        for p in policy_arns:
                            ManagedPolicyArns.append(
                                If(
                                    condname,
                                    ImportValue('IAMPolicy' + p),
                                    Ref('AWS::NoValue')
                                )
                            )

            # resources
            r_Role = IAMRole('IAMRole' + n)
            r_Role.setup(key={'Principal': ''})
            r_Role.Condition = resname + 'RoleAccount'
            r_Role.RoleName = v['UserName']
            r_Role.MaxSessionDuration = 43200
            r_Role.ManagedPolicyArns = ManagedPolicyArns
            AssumeRoleStatement = r_Role.AssumeRolePolicyDocument['Statement'][0]
            AssumeRoleStatement['Condition'] = {
                'StringEquals': {'aws:username': v['UserName']}
            }
            AssumeRoleStatement['Principal'] = {
                'AWS': Sub('arn:aws:iam::${IdAccount}:root', **{'IdAccount': get_endvalue(resname + 'RoleAccount')})
            }

            r_User = IAMUser(resname)
            r_User.setup(key=v, name=n)
            r_User.Groups = RoleGroups

            r_SSMParameter = SSMParameter('SSMParameterPassword' + n)
            r_SSMParameter.Condition = resname
            r_SSMParameter.Name = Sub('/iam/PasswordBase/${UserName}', **{'UserName': get_endvalue(resname + 'UserName')})
            r_SSMParameter.Value = Ref('PasswordBase' + n)
            r_SSMParameter.AllowedPattern = '^[^ ]{16,}$'

            add_obj([
                r_User,
                r_Role,
                r_SSMParameter,
            ])


class IAM_UserToGroupAdditions(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = key + n  # Ex. IAMUserToGroupAdditionBase

            Users = []
            for m, w in v['User'].items():
                condname = '%sUser%s' % (resname, m)
                # conditions
                add_obj(get_condition(condname, 'not_equals', 'None'))

                Users.append(
                    If(
                        condname,
                        # for user defined in the same yaml file
                        # Ref('IAMUser' + m) if m in cfg.IAMUser else get_endvalue(condname),
                        get_endvalue(condname),
                        Ref('AWS::NoValue')
                    )
                )


            # resources
            r_GroupAdd = IAMUserToGroupAddition('IAMUserToGroupAddition' + n)
            r_GroupAdd.setup(key=v, name=n)
            r_GroupAdd.Users = Users

            add_obj([
                r_GroupAdd,
            ])



class IAM_Groups(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = key + n  # Ex. IAMGroupBase

            # conditions
            add_obj(get_condition(resname, 'not_equals', 'None', resname + 'Enabled'))

            ManagedPolicyArns = []
            for m in v['ManagedPolicyArns']:
                if m.startswith('arn'):
                    ManagedPolicyArns.append(m)
                elif m.startswith('Ref('):
                    ManagedPolicyArns.append(eval(m))
                else:
                    ManagedPolicyArns.append(ImportValue('IAMPolicy' + m))


            # resources
            r_Group = IAMGroup(resname)
            r_Group.setup(key=v, name=n)
            r_Group.ManagedPolicyArns = ManagedPolicyArns

            add_obj([
                r_Group,
            ])


class IAM_Policies(object):
    def __init__(self, key):
        # Resources
        for n, v in getattr(cfg, key).items():
            resname = key + n  # Ex. IAMPolicyLambdaR53RecordInstanceId
            Statement = []
            for m, w  in v['Statement'].items():
                Statement.append(IAMPolicyStatement(w))

            if 'Type' in v and v['Type'] == 'Managed':
                r_Policy = IAMManagedPolicy(resname)
            else:
                r_Policy = IAMPolicy(resname)
            r_Policy.setup(key=v, name=n)
            r_Policy.PolicyDocument['Statement'] = Statement

            add_obj(r_Policy)

            # outputs
            if 'Export' in v:
                o_Policy = Output(resname)
                o_Policy.Value = Ref(resname)
                o_Policy.Export = Export(resname)

                add_obj(o_Policy)


class IAM_Roles(object):
    def __init__(self, key):
        # Resources
        for n, v in getattr(cfg, key).items():
            resname = key + n  # Ex. RoleECSService
            r_Role = IAMRole(resname)
            r_Role.setup(key=v)

            # Add embedded policies if present 
            if 'Policies' in v:
                Policies = []
                for p, w in v['Policies'].items():
                    Policies.append(IAMPolicyInRole(p,w))
                r_Role.Policies = Policies

            add_obj(r_Role)

            # outputs
            if 'Export' in v:
                o_Role = Output(resname)
                o_Role.Value = GetAtt(resname, 'Arn')
                o_Role.Export = Export(resname)
                if 'Condition' in v:
                    o_Role.Condition = v['Condition']

                add_obj(o_Role)



import troposphere.ec2 as ec2

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)


# S - SECURITY GROUP #
class SecurityGroup(ec2.SecurityGroup):
    def setup(self):
        self.VpcId = get_expvalue('VpcId')


class SecurityGroupInstanceRules(SecurityGroup):
    def setup(self):
        super(SecurityGroupInstanceRules, self).setup()
        self.GroupDescription = 'Enable Access from LoadBalancer to Instances and between Instances'


class SecurityGroupLoadBalancer(SecurityGroup):
    def setup(self):
        super(SecurityGroupLoadBalancer, self).setup()
        self.GroupDescription = 'Enable access to LoadBalancer'


class SecurityGroupEcsService(SecurityGroup):
    def setup(self):
        super(SecurityGroupEcsService, self).setup()
        self.Condition = 'NetworkModeAwsVpc'
        self.GroupDescription = 'Enable access to Service'
        self.SecurityGroupIngress = []


class SecurityGroupIngress(ec2.SecurityGroupIngress):
    IpProtocol = 'tcp'


class SecurityGroupIngressInstanceELBPorts(SecurityGroupIngress):
    def setup(self, listener):
        name = self.title  # Ex. SecurityGroupIngressListeners1
        self.FromPort = get_endvalue(listener + 'LoadBalancerPort')
        self.ToPort = get_endvalue(listener + 'LoadBalancerPort')
        self.SourceSecurityGroupId = Ref('SecurityGroupLoadBalancer')
        self.GroupId = GetAtt('SecurityGroupInstancesRules', 'GroupId')


class SecurityGroupsIngressEcs(SecurityGroupIngress):
    def setup(self, proto, scheme):
        self.Condition = self.title
        self.GroupId = get_expvalue('SecurityGroupLoadBalancerApplication' + scheme)
        self.FromPort = get_endvalue('ListenerLoadBalancer' + proto + 'Port')
        self.ToPort = self.FromPort

# E - SECURITY GROUP #


# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

class SecurityGroupRule(ec2.SecurityGroupRule):
    IpProtocol = 'tcp'


class SecurityGroupRuleELBPorts(SecurityGroupRule):
    def setup(self):
        name = self.title  # Ex. Listeners1
        self.FromPort = get_endvalue(name + 'LoadBalancerPort')
        self.ToPort = get_endvalue(name + 'LoadBalancerPort')


class SecurityGroupRuleNamePorts(SecurityGroupRule):
    def setup(self):
        name = self.title  # Ex. 3306
        self.FromPort = name
        self.ToPort = name


class SecurityGroupRuleEcsService(SecurityGroupRule):
    def setup(self, scheme):
        self.SourceSecurityGroupId = get_expvalue('SecurityGroupLoadBalancerApplication' + scheme)
        self.FromPort = get_endvalue('ContainerDefinitions1ContainerPort')
        self.ToPort = self.FromPort


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################

class SG_SecurityGroupsExtra(object):
    def __init__(self, Out_String, Out_Map):
        # if key SecurityGroups not present, set and empty SecurityGroups list and output (not added)
        try: cfg.SecurityGroups
        except:
            self.SecurityGroups = []
            self.O_SecurityGroups = Output('') 
            return

        # Parameters
        P_SecurityGroups = Parameter('SecurityGroups')
        P_SecurityGroups.Description = 'SecurityGroups List Extra - %s for default based on env/role' % SECURITY_GROUPS_DEFAULT
        P_SecurityGroups.AllowedPattern = '^(\w*,\w*){%s}$' % (MAX_SECURITY_GROUPS - 1)
        P_SecurityGroups.Default = SECURITY_GROUPS_DEFAULT

        add_obj([
            P_SecurityGroups,
        ])

        SecurityGroups = []

        for n in range(MAX_SECURITY_GROUPS):
            name = 'SecurityGroup%s' % n  # Ex SecurityGroup1
            value = Select(n, Split(',', get_endvalue('SecurityGroups')))  # Ex. ElasticSearch
            outnamename = 'SecurityGroupName%s' % n  # Ex. SecurityGroupName1
            outvaluename = 'SecurityGroupValue%s' % n  # Ex. SecurityGroupValue1

            # conditions
            add_obj(
                {name: Not(
                    get_condition('', 'equals', 'None',
                        Select(n, Split(',', 'SecurityGroups'))
                    )
                )}
            )

            SecurityGroups.append(If(
                name,
                get_expvalue(value, prefix='SecurityGroup'),
                Ref('AWS::NoValue')
            ))

            # outputs
            Out_String.append('${%s}=${%s}' % (outnamename, outvaluename))
            Out_Map.update({
                outnamename: value,
                outvaluename: If(
                    name,
                    get_expvalue(value, prefix='SecurityGroup'),
                    'None'
                )
            })

        # Outputs
        O_SecurityGroups = Output('SecurityGroups')
        O_SecurityGroups.Value = Sub(','.join(Out_String), **Out_Map)

        add_obj([
            O_SecurityGroups,
        ])

        self.O_SecurityGroups = O_SecurityGroups

        self.SecurityGroups = SecurityGroups


class SG_SecurityGroupIngressesExtra(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = key + n
            r_SGI = SecurityGroupIngress(resname)
            auto_get_props(r_SGI, v)

            add_obj(r_SGI)


class SG_SecurityGroupsEC2(object):
    def __init__(self):
        Out_String = ['Rules=${SecurityGroupInstancesRules}']
        Out_Map = {
            'SecurityGroupInstancesRules': {
                'Ref': 'SecurityGroupInstancesRules'
            },
        }


        # Resources
        SG_Extra = SG_SecurityGroupsExtra(Out_String, Out_Map)

        R_SGInstance = SecurityGroupInstanceRules('SecurityGroupInstancesRules')
        R_SGInstance.setup()

        add_obj([
            R_SGInstance,
        ])

        self.SecurityGroups = SG_Extra.SecurityGroups


class SG_SecurityGroupsECS(object):
    def __init__(self):
        Out_String = ['Service=${SecurityGroupEcsService}']
        Out_Map = {
            'SecurityGroupEcsService': {
                'Ref': 'SecurityGroupEcsService'
            },
        }

        # Resources
        SG_Extra = SG_SecurityGroupsExtra(Out_String, Out_Map)

        SG_Extra.O_SecurityGroups.Condition = 'NetworkModeAwsVpc'

        self.SecurityGroups = SG_Extra.SecurityGroups


class SG_SecurityGroupsTSK(object):
    def __init__(self):
        Out_String = []
        Out_Map = {}

        # Resources
        SG_Extra = SG_SecurityGroupsExtra(Out_String, Out_Map)

        SG_Extra.O_SecurityGroups.Condition = 'NetworkModeAwsVpc'

        self.SecurityGroups = SG_Extra.SecurityGroups


def SG_SecurityGroupIngressBaseInstance():
    Securitygroup_Rules = []

    for n, v in cfg.AllowedIp.items():
        name = 'AllowedIp%s' % n
        # conditions
        add_obj(get_condition(name, 'not_equals', 'None', name + 'Enabled'))

        # resources
        Rule = SecurityGroupRule()
        Rule.FromPort = '22'
        Rule.ToPort = '22'
        Rule.CidrIp=get_endvalue(name + 'Ip')

        Securitygroup_Rules.append(
            If(
                name,
                Rule,
                Ref('AWS::NoValue')
            )
        )

    Securitygroup_Rules.append(
        ec2.SecurityGroupRule(
            CidrIp='0.0.0.0/0',
            FromPort='8',
            IpProtocol='icmp',
            ToPort='-1'
        )
    )

    return Securitygroup_Rules


class SG_SecurityGroupRES(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            name = 'SecurityGroup%s' % n  # Ex. SecurityGroupElasticSearchClient
            outname = name if n != 'ElasticSearchClient' else 'SecurityGroup' + n.replace('Client', '')  # Ex. SecurityGroupElasticSearch

            # resources
            r_Base = SecurityGroup(name)
            r_Base.setup()
            r_Base.GroupDescription = get_endvalue('SecurityGroupBase' + n)

            if n == 'BaseInstance':
                r_Base.SecurityGroupIngress = SG_SecurityGroupIngressBaseInstance()

            add_obj(r_Base)

            # outputs
            o_Base = Output(outname)
            o_Base.Value = GetAtt(name, 'GroupId')
            o_Base.Export = Export(outname)

            add_obj(o_Base)


class SG_SecurityGroupIngressesExtraService(object):
    def __init__(self, key, service):  # Ex service=RDS
        Securitygroup_Rules = []
        for n, v in getattr(cfg, key).items():
            name = str(v['FromPort'])  # Ex 3306
            for i in cfg.AllowedIp:
                ipname = 'AllowedIp%s' % i  # Ex. AllowedIp1
                condnameprivate = 'SecurityGroupRulePrivatePort' + name + ipname  # Ex. SecurityGroupRulePrivatePorts3306AllowedIp1

                # conditions
                add_obj([
                    get_condition(ipname, 'not_equals', 'None', ipname + 'Enabled'),
                    {condnameprivate: And(
                        Condition(ipname),
                        get_condition('', 'equals', 'Private', service + 'Access')
                    )},
                ])

                SGRule = SecurityGroupRuleNamePorts(name)
                SGRule.setup()
                SGRule.CidrIp = get_endvalue(ipname + 'Ip')

                Securitygroup_Rules.append(
                    If( 
                        condnameprivate,
                        SGRule,
                        Ref('AWS::NoValue'),
                    )
                )

            SGRule = SecurityGroupRuleNamePorts(name)
            SGRule.setup()
            SGRule.CidrIp = '0.0.0.0/0'

            Securitygroup_Rules.append(
                If(
                    service + 'Public',
                    SGRule,
                    Ref('AWS::NoValue'),
                )
            )

        # Resources
        SG_SecurityGroupIngressesExtra(key=key)

        R_SG = SecurityGroup('SecurityGroup' + service)
        R_SG.setup()
        R_SG.GroupDescription = 'Enable access to ' + service
        R_SG.SecurityGroupIngress = Securitygroup_Rules

        add_obj([
            R_SG,
        ])

        # Outputs
        O_SG = Output('SecurityGroup')
        O_SG.Value = GetAtt('SecurityGroup' + service, 'GroupId')

        add_obj([
            O_SG,
        ])

class SG_SecurityGroupIngressesExtraRDS(SG_SecurityGroupIngressesExtraService):
    def __init__(self, **kwargs):
        super(SG_SecurityGroupIngressesExtraRDS, self).__init__(service='RDS', **kwargs)


class SG_SecurityGroupIngressesExtraCCH(SG_SecurityGroupIngressesExtraService):
    def __init__(self, **kwargs):
        super(SG_SecurityGroupIngressesExtraCCH, self).__init__(service='CCH', **kwargs)



from collections import OrderedDict, Mapping
import troposphere.apigateway as agw

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)
from .lambdas import LambdaPermissionApiGateway


class ApiGatewayAccount(agw.Account):
    def setup(self):
        self.CloudWatchRoleArn = get_expvalue('RoleApiGatewayCloudWatch')


class ApiGatewayResource(agw.Resource):
    def setup(self, key, stage):
        mapname = 'ApiGatewayStage' + stage + 'ApiGatewayResource'
        stagekey = cfg.ApiGatewayStage[stage]['ApiGatewayResource']
        auto_get_props(self, key)
        self.RestApiId = Ref('ApiGatewayRestApi')
        auto_get_props(self, stagekey, mapname=mapname, recurse=True, rootkey=key, rootname=self.title)


class ApiGatewayMethod(agw.Method):
    def setup(self, key, basename, name, stage):
        mapname = 'ApiGatewayStage' + stage + 'ApiGatewayResource' + basename + 'Method' + name
        stagekey = cfg.ApiGatewayStage[stage]['ApiGatewayResource'][basename]['Method'][name] 
        auto_get_props(self, key, recurse=True)
        self.RestApiId = Ref('ApiGatewayRestApi')
        self.ResourceId = Ref('ApiGatewayResource' + basename)
        auto_get_props(self, stagekey, mapname=mapname, recurse=True, rootkey=key, rootname=self.title)
        # If Uri is a lambda get same stage version lambda name Ex. LambdaPaidApiv1
        if ':lambda:' in self.Integration.Uri:
            st = self.Integration.Uri
            l = st.rfind('.')
            self.Integration.Uri = Sub(st[0:l] + stage + st[l:])


class ApiGatewayStage(agw.Stage):
    def setup(self, name, key):
        self.StageName = name
        auto_get_props(self, key, recurse=True)
        self.RestApiId = Ref('ApiGatewayRestApi')
        self.DeploymentId = Ref('ApiGatewayDeployment' + name)


class ApiGatewayDeployment(agw.Deployment):
    def setup(self, name, key):
        lastresource = next(reversed(cfg.ApiGatewayResource))
        lastmethod = next(reversed(cfg.ApiGatewayResource[lastresource]['Method']))
        self.DependsOn = 'ApiGatewayResource' + lastresource + 'Method' + lastmethod
        self.Description = Ref('Deployment' + name + 'Description')
        self.RestApiId = Ref('ApiGatewayRestApi')


class AGW_Account(object):
    def __init__(self, key):
        # Resources
        R_Account = ApiGatewayAccount('ApiGatewayAccount')
        R_Account.setup()

        add_obj([
            R_Account,
        ])


class AGW_DomainName(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = key + n
            # resources
            r_Domain = agw.DomainName(resname)
            auto_get_props(r_Domain, v, recurse=True)

            # outputs
            o_Domain = Output(resname)
            o_Domain.Value = Ref(resname)
            o_Domain.Export = Export(resname)

            add_obj([
                r_Domain,
                o_Domain,
            ])


class AGW_UsagePlans(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = key + n
            for m, w in v['ApiStages'].items():
                # parameters
                p_Stage = Parameter(resname + 'ApiStages' + m + 'Stage')
                p_Stage.Description = m + ' Stage - empty for default based on env/role'

                add_obj(p_Stage)

            # resources
            r_UsagePlan = agw.UsagePlan(resname)
            auto_get_props(r_UsagePlan, v, recurse=True)

            add_obj([
                r_UsagePlan,
            ])

            # outputs
            o_UsagePlan = Output(resname)
            o_UsagePlan.Value = Ref(resname)
            o_UsagePlan.Export = Export(resname)

            add_obj([
                o_UsagePlan,
            ])


class AGW_ApiKeys(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = key + n
            # parameters
            p_Enabled = Parameter(resname + 'Enabled')
            p_Enabled.Description = resname + 'Enabled - empty for default based on env/role'

            p_UsagePlan = Parameter(resname + 'UsagePlan')
            p_UsagePlan.Description = resname + 'UsagePlan - empty for default based on env/role'

            add_obj([
                p_Enabled,
                p_UsagePlan,
            ])
            
            # resources
            r_ApiKey = agw.ApiKey(resname)
            auto_get_props(r_ApiKey, v, recurse=True)

            add_obj([
                r_ApiKey,
            ])

            if 'UsagePlan' in v:
                plankey_name = resname + 'UsagePlan'
                r_UsagePlanKey = agw.UsagePlanKey('ApiGatewayUsagePlan' + n)
                r_UsagePlanKey.KeyId = Ref(resname)
                r_UsagePlanKey.KeyType = 'API_KEY'
                r_UsagePlanKey.UsagePlanId = ImportValue(get_subvalue('ApiGatewayUsagePlan${1M}', resname + 'UsagePlan'))

                add_obj(r_UsagePlanKey)
            
            # outputs
            o_ApiKey = Output(resname)
            o_ApiKey.Value = Ref(resname)


class AGW_Stages(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            # parameters
            p_DeploymentDescription = Parameter('Deployment' + n + 'Description')
            p_DeploymentDescription.Description = 'Deployment' + n + ' Description'
            p_DeploymentDescription.Default = n

            add_obj([
                p_DeploymentDescription,
            ])

            # resources
            resname = key + n
            r_Stage = ApiGatewayStage(resname)
            r_Stage.setup(name=n, key=v)

            r_Deployment = ApiGatewayDeployment('ApiGatewayDeployment' + n)
            r_Deployment.setup(name=n, key=v)

            add_obj([
                r_Stage,
                r_Deployment,
            ])
       

class AGW_RestApi(object):
    def __init__(self, key):
        # Resources
        R_RestApi = agw.RestApi('ApiGatewayRestApi')
        auto_get_props(R_RestApi, mapname='')
        R_RestApi.EndpointConfiguration = agw.EndpointConfiguration(
            Types=get_endvalue('EndpointConfiguration')
        )

        add_obj([
            R_RestApi,
        ])

        for n, v in getattr(cfg, key).items():
            resname = key + n
            agw_stage = cfg.Stage
            r_Resource = ApiGatewayResource(resname)
            r_Resource.setup(key=v, stage=agw_stage)

            for m, w in v['Method'].items():
                r_Method = ApiGatewayMethod(resname + 'Method' + m)
                r_Method.setup(key=w, basename=n, name=m, stage=agw_stage)

                add_obj([
                    r_Resource,
                    r_Method,
                ])

        for l, w in cfg.Lambda.items():
            r_LambdaPermission = LambdaPermissionApiGateway('LambdaPermission' + l)
            r_LambdaPermission.setup(
                name=Ref('Lambda' + l),
                source=Sub('arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:${ApiGatewayRestApi}/*/*/*')
            )

            add_obj(r_LambdaPermission)



import os
import sys

import troposphere.autoscaling as asg
import troposphere.cloudformation as cfm
import troposphere.policies as pol
import troposphere.applicationautoscaling as aas

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
    get_subvalue, auto_get_props, get_condition, add_obj)
from .cfn import *
from .codedeploy import CD_DeploymentGroup
from .securitygroup import SG_SecurityGroupsEC2
from .iam import IAMInstanceProfile

parent_dir_name = os.getcwd()
sys.path.append(parent_dir_name + '/lib')

try:
    from cfnExt import *
except ImportError:
    pass


# S - AUTOSCALING #
class ASAutoScalingGroup(asg.AutoScalingGroup):
    def setup(self, spot=None):
        if spot:
            CapacityDesiredASGMainIsSpot = get_endvalue('CapacityDesired')
            CapacityDesiredASGMainIsNotSpot = 0
            CapacityMinASGMainIsSpot = get_endvalue('CapacityMin')
            CapacityMinASGMainIsNotSpot = 0
            self.Condition = 'SpotASG'
            self.LaunchConfigurationName = Ref('LaunchConfigurationSpot')
            self.UpdatePolicy = ASUpdatePolicy(spot=True)
        else:
            CapacityDesiredASGMainIsSpot = 0
            CapacityDesiredASGMainIsNotSpot = get_endvalue('CapacityDesired')
            CapacityMinASGMainIsSpot = 0
            CapacityMinASGMainIsNotSpot = get_endvalue('CapacityMin')
            self.LaunchConfigurationName = Ref('LaunchConfiguration')
            self.UpdatePolicy = ASUpdatePolicy()

        self.AvailabilityZones = GetAZs()

        if cfg.SpotASG:
            self.DesiredCapacity = If('ASGMainIsSpot', CapacityDesiredASGMainIsSpot, CapacityDesiredASGMainIsNotSpot)
            self.MinSize = If('ASGMainIsSpot', CapacityMinASGMainIsSpot, CapacityMinASGMainIsNotSpot)
        else:
            self.DesiredCapacity = get_endvalue('CapacityDesired')
            self.MinSize = get_endvalue('CapacityMin')

        self.CreationPolicy = pol.CreationPolicy(
            ResourceSignal=pol.ResourceSignal(
                Count=self.DesiredCapacity,
                Timeout=get_endvalue('AutoscalingCreationTimeout')
            )
        )
        self.HealthCheckGracePeriod = get_endvalue('HealthCheckGracePeriod')
        self.HealthCheckType = get_endvalue('HealthCheckType')
        self.MaxSize = get_endvalue('CapacityMax')
        self.MetricsCollection = [asg.MetricsCollection(
            Granularity='1Minute'
        )]
        self.Tags = [
            asg.Tag(('Name'), Ref('EnvRole'), True),
            asg.Tag(('EnvStackName'), Ref('AWS::StackName'), True),
        ]
        self.TerminationPolicies = ['OldestInstance']
        self.VPCZoneIdentifier = Split(',', get_expvalue('SubnetsPrivate'))


class ASLaunchConfiguration(asg.LaunchConfiguration):
    def setup(self, UserDataApp, spot=None):
        if spot:
            AutoScalingGroupName='AutoScalingGroupSpot'
            self.Condition = 'SpotASG'
        else:
            AutoScalingGroupName = 'AutoScalingGroup'
        self.AssociatePublicIpAddress = False
        self.BlockDeviceMappings = [
            asg.BlockDeviceMapping(
                DeviceName='/dev/xvda',
                Ebs=asg.EBSBlockDevice(
                    VolumeSize=get_endvalue('VolumeSize'),
                    VolumeType=get_endvalue('VolumeType'),
                )
            ),
            If(
                'AdditionalStorage',
                asg.BlockDeviceMapping(
                    DeviceName=get_endvalue('AdditionalStorageName'),
                    Ebs=asg.EBSBlockDevice(
                        VolumeSize=get_endvalue('AdditionalStorageSize'),
                        VolumeType=get_endvalue('AdditionalStorageType'),
                    )
                ),
                Ref('AWS::NoValue')
            ),
            If(
                'InstaceEphemeral0',
                asg.BlockDeviceMapping(
                    DeviceName='/dev/xvdb',
                    VirtualName='ephemeral0'
                ),
                Ref('AWS::NoValue')
            ),
            If(
                'InstaceEphemeral1',
                asg.BlockDeviceMapping(
                    DeviceName='/dev/xvdc',
                    VirtualName='ephemeral1'
                ),
                Ref('AWS::NoValue')
            ),
            If(
                'InstaceEphemeral2',
                asg.BlockDeviceMapping(
                    DeviceName='/dev/xvdd',
                    VirtualName='ephemeral2'
                ),
                Ref('AWS::NoValue')
            ),
        ]
        self.IamInstanceProfile = Ref('InstanceProfile')
        self.ImageId = If(
            'ImageIdLatest',
            Ref('ImageIdLatest'),
            get_endvalue('ImageId'),
        ) if 'ImageIdLatest' in cfg.Parameter else get_endvalue('ImageId')
        self.InstanceMonitoring = get_endvalue('InstanceMonitoring')
        self.InstanceType = get_endvalue('InstanceType')
        self.KeyName = get_endvalue('KeyName')
        self.SecurityGroups = [
            GetAtt('SecurityGroupInstancesRules', 'GroupId'),
        ]
        self.SpotPrice = If('SpotPrice', get_endvalue('SpotPrice'), Ref('AWS::NoValue'))
        self.UserData = Base64(Join('', [
            '#!/bin/bash\n',
            'PATH=/opt/aws/bin:$PATH\n',
            'export BASH_ENV=/etc/profile.d/ibox_env.sh\n',
            'export ENV=$BASH_ENV\n',
            'yum -C list installed aws-cfn-bootstrap || yum install -y aws-cfn-bootstrap\n',
            Sub(''.join(UserDataApp)),
            'cfn-init -v',
            ' --stack ', Ref('AWS::StackName'),
            ' --role ', Ref('RoleInstance'),
            ' --resource LaunchConfiguration',
            ' --region ', Ref('AWS::Region'), '\n',
            If(
                'DoNotSignal',
                Ref('AWS::NoValue'),
                Sub('cfn-signal -e $? --stack ${AWS::StackName} --role ${RoleInstance} --resource %s --region ${AWS::Region}\n' % AutoScalingGroupName)
            ),
            'rm /var/lib/cloud/instance/sem/config_scripts_user\n',
        ]))


class ASScalingPolicyStep(asg.ScalingPolicy):
    def setup(self):
        name = self.title  # Ex. ScalingPolicyDown
        self.AdjustmentType = 'ChangeInCapacity'
        self.AutoScalingGroupName = Ref('AutoScalingGroup')
        self.EstimatedInstanceWarmup = get_endvalue(name + 'EstimatedInstanceWarmup')
        self.PolicyType = 'StepScaling'
        self.StepAdjustments = [
            asg.StepAdjustments(
                MetricIntervalLowerBound=get_endvalue(name + 'MetricIntervalLowerBound1'),  # Ex. ScalingPolicyDownMetricIntervalLowerBound1
                MetricIntervalUpperBound=get_endvalue(name + 'MetricIntervalUpperBound1'),
                ScalingAdjustment=get_endvalue(name + 'ScalingAdjustment1')  # Ex. ScalingPolicyDownScalingAdjustment1
            ),
            asg.StepAdjustments(
                MetricIntervalLowerBound=get_endvalue(name + 'MetricIntervalLowerBound2'),
                MetricIntervalUpperBound=get_endvalue(name + 'MetricIntervalUpperBound2'),
                ScalingAdjustment=get_endvalue(name + 'ScalingAdjustment2')
            ),
            asg.StepAdjustments(
                ScalingAdjustment=get_endvalue(name + 'ScalingAdjustment3')
            )
        ]


class ASScheduledAction(asg.ScheduledAction):
    def setup(self):
        name = self.title
        self.Condition = name

        if cfg.SpotASG:
            self.AutoScalingGroupName = If('ASGMainIsSpot', Ref('AutoScalingGroupSpot'), Ref('AutoScalingGroup'))
        else:
            self.AutoScalingGroupName = Ref('AutoScalingGroup')

        self.DesiredCapacity = If(
            name + 'CapacityDesiredSize',
            get_endvalue('CapacityDesired'),
            get_endvalue(name + 'DesiredSize', nocondition=name + 'KeepDesiredSize')
        )
        self.MinSize = If(
            name + 'CapacityMinSize',
            get_endvalue('CapacityMin'),
            get_endvalue(name + 'MinSize', nocondition=name + 'KeepMinSize')
        )
        self.MaxSize = If(
            name + 'CapacityMaxSize',
            get_endvalue('CapacityMax'),
            get_endvalue(name + 'MaxSize', nocondition=name + 'KeepMaxSize')
        )
        self.Recurrence = get_endvalue(name + 'Recurrence')
        self.StartTime = If(
            name + 'StartTimeOverride',
            Ref(name + 'StartTime'),
            Ref('AWS::NoValue')
        )


class ASLifecycleHook(asg.LifecycleHook):
    def __init__(self, title, name, key, **kwargs):
        super(asg.LifecycleHook, self).__init__(title, **kwargs)
        auto_get_props(self, key)
        self.HeartbeatTimeout = get_endvalue(title + 'HeartbeatTimeout')

# E - AUTOSCALING #


class ASScalingPolicyStepDown(ASScalingPolicyStep):
    def setup(self):
        super(ASScalingPolicyStepDown, self).setup()
        name = self.title
        self.StepAdjustments[2].MetricIntervalUpperBound = get_endvalue(name + 'MetricIntervalUpperBound3')


class ASScalingPolicyStepUp(ASScalingPolicyStep):
    def setup(self):
        super(ASScalingPolicyStepUp, self).setup()
        name = self.title
        self.StepAdjustments[2].MetricIntervalLowerBound = get_endvalue(name + 'MetricIntervalLowerBound3')


# S - APPLICATION AUTOSCALING #
class APPASScalingPolicy(aas.ScalingPolicy):
    def setup(self):
        name = self.title
        self.PolicyName = self.title
        self.ScalingTargetId = Ref('ScalableTarget')


class APPASScalingPolicyStep(APPASScalingPolicy):
    def setup(self):
        super(APPASScalingPolicyStep, self).setup()
        name = self.title
        self.PolicyType = 'StepScaling'
        self.StepScalingPolicyConfiguration = aas.StepScalingPolicyConfiguration(
            AdjustmentType='ChangeInCapacity',
            MetricAggregationType='Average',
            StepAdjustments=[
                aas.StepAdjustment(
                    MetricIntervalLowerBound=get_endvalue(name + 'MetricIntervalLowerBound1'),
                    MetricIntervalUpperBound=get_endvalue(name + 'MetricIntervalUpperBound1'),
                    ScalingAdjustment=get_endvalue(name + 'ScalingAdjustment1')
                    ),
                aas.StepAdjustment(
                    MetricIntervalLowerBound=get_endvalue(name + 'MetricIntervalLowerBound2'),
                    MetricIntervalUpperBound=get_endvalue(name + 'MetricIntervalUpperBound2'),
                    ScalingAdjustment=get_endvalue(name + 'ScalingAdjustment2')
                    ),
                aas.StepAdjustment(
                    ScalingAdjustment=get_endvalue(name + 'ScalingAdjustment3')
                )
            ]
        )


class APPASScalingPolicyStepDown(APPASScalingPolicyStep):
    def setup(self):
        super(APPASScalingPolicyStepDown, self).setup()
        name = self.title
        self.StepScalingPolicyConfiguration.StepAdjustments[2].MetricIntervalUpperBound = get_endvalue(name + 'MetricIntervalUpperBound3')


class APPASScalingPolicyStepUp(APPASScalingPolicyStep):
    def setup(self):
        super(APPASScalingPolicyStepUp, self).setup()
        name = self.title
        self.StepScalingPolicyConfiguration.StepAdjustments[2].MetricIntervalLowerBound = get_endvalue(name + 'MetricIntervalLowerBound3')


class APPASScheduledAction(aas.ScheduledAction):
    def setup(self):
        name = self.title
        self.ScalableTargetAction = aas.ScalableTargetAction(
            MinCapacity=If(
                name + 'CapacityMinSize',
                get_endvalue('CapacityMin'),
                get_endvalue(name + 'MinSize', nocondition=name + 'KeepMinSize')
            ),
            MaxCapacity=If(
                name + 'CapacityMaxSize',
                get_endvalue('CapacityMax'),
                get_endvalue(name + 'MaxSize', nocondition=name + 'KeepMaxSize')
            )
        )
        self.Schedule = get_endvalue(name + 'Recurrence')
        self.ScheduledActionName = name
        self.StartTime = If(
            name + 'StartTimeOverride',
            Ref(name + 'StartTime'),
            Ref('AWS::NoValue')
        )


class APPASScalableTarget(aas.ScalableTarget):
    def setup(self):
        self.MaxCapacity = get_endvalue('CapacityMax')
        self.MinCapacity = get_endvalue('CapacityMin')
        self.ResourceId = get_subvalue('service/${1E}/${Service.Name}', 'Cluster', 'ClusterStack')
        self.RoleARN = get_expvalue('RoleEC2ContainerServiceAutoscale', '')
        self.ScalableDimension = 'ecs:service:DesiredCount'
        self.ServiceNamespace = 'ecs'

# E - APPLICATION AUTOSCALING #

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

# S - AUTOSCALING #
class ASUpdatePolicy(pol.UpdatePolicy):
    def __init__(self, spot=None, **kwargs):
        super(ASUpdatePolicy, self).__init__(**kwargs)
        if spot:
            WillReplaceCondition = 'WillReplaceSpot' 
            AutoScalingRollingUpdateCondition = 'RollingUpdateSpot' 
        else:
            WillReplaceCondition = 'WillReplace' 
            AutoScalingRollingUpdateCondition = 'RollingUpdate' 
        self.AutoScalingReplacingUpdate = If(
            WillReplaceCondition,
            pol.AutoScalingReplacingUpdate(
                WillReplace=True
            ),
            Ref('AWS::NoValue')
        )
        self.AutoScalingRollingUpdate = If(
            AutoScalingRollingUpdateCondition,
            pol.AutoScalingRollingUpdate(
                MaxBatchSize=get_endvalue('RollingUpdateMaxBatchSize'),
                # MinInstancesInService=get_endvalue('RollingUpdateMinInstancesInService'),
                MinInstancesInService=get_endvalue('CapacityDesired'),
                MinSuccessfulInstancesPercent=get_endvalue('RollingUpdateMinSuccessfulInstancesPercent'),
                PauseTime=get_endvalue('RollingUpdatePauseTime'),
                SuspendProcesses=[
                    'HealthCheck',
                    'ReplaceUnhealthy',
                    'AlarmNotification',
                    'ScheduledActions'
                ],
                WaitOnResourceSignals=True
            ),
            Ref('AWS::NoValue')
        )
        self.AutoScalingScheduledAction=pol.AutoScalingScheduledAction(
            IgnoreUnmodifiedGroupSizeProperties=True
        )


class ASInitConfigSets(cfm.InitConfigSets):
    def setup(self):
        if cfg.Apps:
            CODEDEPLOY = If('DeploymentGroup', 'CODEDEPLOY', Ref('AWS::NoValue'))
        else:
            CODEDEPLOY = Ref('AWS::NoValue')

        if cfg.LoadBalancerClassic or cfg.LoadBalancerApplication:
            ELBWAITER = 'ELBWAITER'
        else:
            ELBWAITER = Ref('AWS::NoValue')

        self.data = {
            'default': [
                'REPOSITORIES',
                'PACKAGES',
                'SETUP',
                CODEDEPLOY,
                'SERVICES',
                ELBWAITER,
            ],
            'buildami': [
                'REPOSITORIES',
                'PACKAGES',
            ],
            'buildamifull': [
                'REPOSITORIES',
                'PACKAGES',
                'SETUP',
            ]
        }


class ASInitConfigSetup(cfm.InitConfig):
    def __init__(self, **kwargs):
        self.ibox_env_app = ''
        super(ASInitConfigSetup, self).__init__(**kwargs)

    def setup(self):
        self.files = {
            '/etc/profile.d/ibox_env.sh': {
                'content': Join('', [
                    '#setup ibox environment variables\n',
                    'export Env=', Ref('Env'), '\n',
                    'export EnvAbbr=', Ref('EnvShort'), '\n',
                    'export EnvRegion=', Ref('AWS::Region'), '\n',
                    'export EnvAccountId=', Ref('AWS::AccountId'), '\n',
                    'export EnvRole=', Ref('EnvRole'), '\n',
                    'export EnvBrand=', cfg.BrandDomain, '\n',
                    'export EnvStackName=', Ref('AWS::StackName'), '\n',
                ] + self.ibox_env_app)
            },
            '/etc/cfn/cfn-hup.conf': {
                'content': Join('', [
                    '[main]\n',
                    'stack=', Ref('AWS::StackId'), '\n',
                    'region=', Ref('AWS::Region'), '\n',
                    'role=', Ref('RoleInstance'), '\n',
                    'interval=5\n'
                ]),
                'mode': '000400',
                'owner': 'root',
                'group': 'root'
            },
            '/etc/cfn/hooks.d/cfn-auto-reloader.conf': {
                'content': Join('', [
                    '[cfn-auto-reloader-hook]\n',
                    'triggers=post.add, post.update\n',
                    'path=Resources.LaunchConfiguration.Metadata.CloudFormationInitVersion\n',
                    'action=/opt/aws/bin/cfn-init -v',
                    ' --stack ', Ref('AWS::StackName'),
                    ' --role ', Ref('RoleInstance'),
                    ' --resource LaunchConfiguration',
                    ' --region ', Ref('AWS::Region'), '\n',
                    'runas=root\n'
                ])
            },
            '/usr/local/bin/chamber': {
                'mode': '000755',
                'source': Sub('https://%s.s3.${AWS::Region}.amazonaws.com/ibox-tools/chamber' % cfg.BucketAppRepository),
                'owner': 'root',
                'group': 'root',
            },
        }
        self.commands = {
            '01_disk': {
                'command': Join('', [
                    'n=0\n',
                    'for disk in /dev/xvd[b-d]; do\n',
                    '  [ -b "$disk" ] || continue\n',
                    '  file -s "$disk" | grep -q "ext[34] filesystem" || ',
                    '  { mkfs.ext4 $disk || continue; }\n',
                    '  mkdir -p /media/ephemeral${n} && ',
                    '  mount $disk /media/ephemeral${n}\n',
                    '  n=$(($n+1))\n',
                    'done'
                ])
            },
            '02_disk_additional': If(
                'AdditionalStorageMount',
                {
                    #'command': Join('', [
                    #    'file -s ', get_endvalue('AdditionalStorageName'), '1', ' | grep -q "ext[34] filesystem" || ',
                    #    '{ parted -s ', get_endvalue('AdditionalStorageName'), ' mklabel gpt', ' && ',
                    #    'parted -s ', get_endvalue('AdditionalStorageName'), ' mkpart primary ext2 1 ',
                    #    get_endvalue('AdditionalStorageSize'), 'G', ' && ',
                    #    'mkfs.ext4 ', get_endvalue('AdditionalStorageName'), '1', ' || continue; }\n',
                    #    'mkdir -p /data', ' && ',
                    #    'mount /dev/xvdx1 /data'
                    #])
                    'command': get_subvalue(
                        'file -s ${1M}1 | grep -q "ext[34] filesystem" || { parted -s ${1M} mklabel gpt && parted -s ${1M} mkpart primary ext2 1 ${2M}G && mkfs.ext4 ${1M}1 || continue; }\nmkdir -p /data && mount ${1M}1 /data',
                        ['AdditionalStorageName', 'AdditionalStorageSize']
                    )
                },
                Ref('AWS::NoValue')
            ),
            '03_efs_mounts': If(
                'EfsMounts',
                {
                    'command': Join('', [
                        'for mounts in ', Join(' ', Ref('EfsMounts')), ';do\n',
                        '  mkdir -p "/media/${mounts}"\n',
                        '  mountpoint -q "/media/${mounts}" ||',
                        '    mount -t nfs4 -o nfsvers=4,rsize=1048576,wsize=1048576,hard,timeo=600,retrans=2 ',
                        '    efs-${mounts}.', cfg.HostedZoneNamePrivate, ':/ ',
                        '    /media/${mounts}\n',
                        'done'
                    ]),
                },
                Ref('AWS::NoValue')
            ),
            '04_rmdir_tmp_ibox': {
                'command': 'rm -fr /tmp/ibox'
            }
        }
        self.services = {
            'sysvinit': {
                'cfn-hup': {
                    'enabled': 'false',
                    'ensureRunning': 'true',
                    'files': [
                        '/etc/cfn/cfn-hup.conf',
                        '/etc/cfn/hooks.d/cfn-auto-reloader.conf'
                    ]
                }
            }
        }


class ASInitConfigCodeDeploy(cfm.InitConfig):
    def setup(self):
        self.files = {
            '/etc/codedeploy-agent/conf/codedeployagent.yml': {
                'content': Join('', [
                    '---\n',
                    ':log_aws_wire: false\n',
                    ':log_dir: \'/var/log/aws/codedeploy-agent/\'\n',
                    ':pid_dir: \'/opt/codedeploy-agent/state/.pid/\'\n',
                    ':program_name: codedeploy-agent\n',
                    ':root_dir: \'/opt/codedeploy-agent/deployment-root\'\n',
                    ':verbose: false\n',
                    ':wait_between_runs: 1\n',
                    ':proxy_uri:\n',
                    ':max_revisions: 2\n'
                ])
            },
            '/tmp/codedeployinstall.sh': {
                'source': Sub('https://aws-codedeploy-${AWS::Region}.s3.amazonaws.com/latest/install'),
                'mode': '000700'
            }
        }
        self.commands = {
            '03_codedeploy-install-and-run': {
                'command': '/tmp/codedeployinstall.sh auto'
            }
        }
        self.services = {
            'sysvinit': {
                'codedeploy-agent': {
                    'enabled': 'false',
                    'ensureRunning': 'true',
                    'files': ['/etc/codedeploy-agent/conf/codedeployagent.yml']
                }
            }
        }


class ASInitConfigApps(cfm.InitConfig):
    def __init__(self, title, **kwargs):
        super(ASInitConfigApps, self).__init__(title, **kwargs)
        name = self.title  # Ex. Apps1
        reponame = name + 'RepoName'
        n = name.replace('Apps', '')
        envappversion = 'EnvApp%sVersion' % n

        self.sources = {
            '/tmp/ibox/': If(
                'DeployRevision',
                Ref('AWS::NoValue'),
                get_subvalue(
                    'https://%s.s3-${AWS::Region}.amazonaws.com/${1M}/${1M}-${%s}.tar.gz' % (cfg.BucketAppRepository, envappversion),
                    reponame, ''
                )
            )
        }
        self.commands = {
            '01_setup': If(
                'DeployRevision',
                Ref('AWS::NoValue'),
                {
                    'command': get_subvalue(
                        'EnvAppVersion=${%s} EnvRepoName=${1M} /tmp/ibox/bin/setup.sh' % envappversion, reponame
                    )
                }
            ),
            '02_setup_reboot_codedeploy': If(
                'DeployRevision',
                {
                    'command': get_subvalue(
                        'EnvAppVersion=${%s} EnvRepoName=${1M} /opt/ibox/${1M}/live/bin/setup.sh' % envappversion, reponame
                    ),
                    'test': 'test -e /var/lib/cloud/instance/sem/config_ssh_authkey_fingerprints'
                },
                Ref('AWS::NoValue')
            ),
            '03_rmdir_tmp_ibox': {
                'command': 'rm -fr /tmp/ibox'
            }
        }


class ASInitConfigAppsBuildAmi(ASInitConfigApps):
    def __init__(self, title, **kwargs):
        super(ASInitConfigAppsBuildAmi, self).__init__(title, **kwargs)
        for n, v in self.sources.items():
            if not isinstance(v, dict):
                if 'Fn::If' in v.data:
                    self.sources[n] = v.data['Fn::If'][2]

        for n, v in self.commands.items():
            if not isinstance(v, dict):
                if 'Fn::If' in v.data:
                    self.commands[n] = v.data['Fn::If'][2]


class ASInitConfigELBClassicExternal(cfm.InitConfig):
    def setup(self):
        self.commands = {
            'ELBClassicExternalHealthCheck': {
                'command': Join('', [
                    'until [ "$state" = \'"InService"\' ]; do',
                    '  state=$(aws --region ', Ref('AWS::Region'), ' elb describe-instance-health',
                    '  --load-balancer-name ', Ref('LoadBalancerClassicExternal'),
                    '  --instances $(curl -s http://169.254.169.254/latest/meta-data/instance-id)',
                    '  --query InstanceStates[0].State);',
                    '  sleep 10;',
                    'done'
                ])
            }
        }


class ASInitConfigELBClassicInternal(cfm.InitConfig):
    def setup(self):
        self.commands = {
            'ELBClassicInternalHealthCheck': {
                'command': Join('', [
                    'until [ "$state" = \'"InService"\' ]; do',
                    '  state=$(aws --region ', Ref('AWS::Region'), ' elb describe-instance-health',
                    '  --load-balancer-name ', Ref('LoadBalancerClassicInternal'),
                    '  --instances $(curl -s http://169.254.169.254/latest/meta-data/instance-id)',
                    '  --query InstanceStates[0].State);',
                    '  sleep 10;',
                    'done'
                ])
            }
        }


class ASInitConfigELBApplicationExternal(cfm.InitConfig):
    def setup(self):
        self.commands = {
            'ELBApplicationExternalHealthCheck': {
                'command': Join('', [
                    'until [ "$state" = \'"healthy"\' ]; do',
                    '  state=$(aws --region ', Ref('AWS::Region'), ' elbv2 describe-target-health',
                    '  --target-group-arn ', Ref('TargetGroupExternal'),
                    '  --targets Id=$(curl -s http://169.254.169.254/latest/meta-data/instance-id)',
                    '  --query TargetHealthDescriptions[0].TargetHealth.State);',
                    '  sleep 10;',
                    'done'
                ])
            }
        }


class ASInitConfigELBApplicationInternal(cfm.InitConfig):
    def setup(self):
        self.commands = {
            'ELBApplicationInternalHealthCheck': {
                'command': Join('', [
                    'until [ "$state" = \'"healthy"\' ]; do',
                    '  state=$(aws --region ', Ref('AWS::Region'), ' elbv2 describe-target-health',
                    '  --target-group-arn ', Ref('TargetGroupInternal'),
                    '  --targets Id=$(curl -s http://169.254.169.254/latest/meta-data/instance-id)',
                    '  --query TargetHealthDescriptions[0].TargetHealth.State);',
                    '  sleep 10;',
                    'done'
                ])
            }
        }


class ASNotificationConfiguration(asg.NotificationConfigurations):
    def setup(self):
        self.NotificationTypes = [
            'autoscaling:EC2_INSTANCE_LAUNCH',
            'autoscaling:EC2_INSTANCE_TERMINATE',
        ]
        self.TopicARN = get_expvalue('SNSTopicASGNotification')
# E - AUTOSCALING #

# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################

class AS_ScheduledAction(object):
    def __init__(self, resname, OutKey=[]):
        OutKey.extend([
            'MinSize',
            'MaxSize',
            'Recurrence',
            'StartTime',
        ])
        # parameters
        P_MinSize = Parameter(resname + 'MinSize')
        P_MinSize.Description = resname + 'Min Capacity - k to keep current value - empty for default based on env/role'

        P_MaxSize = Parameter(resname + 'MaxSize')
        P_MaxSize.Description = resname + 'Max Capacity - k to keep current value - empty for default based on env/role'

        P_Recurrence = Parameter(resname + 'Recurrence')
        P_Recurrence.Description = resname + 'Recurrence - k to keep current value - empty for default based on env/role'

        P_StartTime = Parameter(resname + 'StartTime')
        P_StartTime.Description = resname + 'StartTime - k to keep current value - empty for default based on env/role'

        add_obj([
            P_MinSize,
            P_MaxSize,
            P_Recurrence,
            P_StartTime,
        ])

        # conditions
        add_obj([
            get_condition(resname + 'KeepMinSize', 'equals', 'k', resname + 'MinSize'),
            get_condition(resname + 'KeepMaxSize', 'equals', 'k', resname + 'MaxSize'),
            get_condition(resname + 'CapacityMinSize', 'equals', 'CapacityMin', resname + 'MinSize'),
            get_condition(resname + 'CapacityMaxSize', 'equals', 'CapacityMax', resname + 'MaxSize'),
            get_condition(resname, 'not_equals', 'None', resname + 'Recurrence'),
        ])

        # outputs
        out_String = []
        out_Map = {}
        for k in OutKey:
            out_String.append('%s=${%s}' % (k, k))  # Ex. 'MinSize=${MinSize}'
            out_Map.update({k: get_endvalue(resname + k) if k != 'StartTime' else Ref(resname + k)})

        O_ScheduledAction = Output(resname)
        O_ScheduledAction.Value = Sub(','.join(out_String), **out_Map)
        
        add_obj(O_ScheduledAction)


class AS_ScheduledActionsEC2(object):
    def __init__(self, key):
        for n, v in getattr(cfg, key).items():
            resname = '%s%s' % (key, n)
            # parameters
            p_DesiredSize = Parameter(resname + 'DesiredSize')
            p_DesiredSize.Description = resname + 'Desired Capacity - k to keep current value - empty for default based on env/role'
    
            add_obj(p_DesiredSize)
    
            # conditions
            add_obj([
                get_condition(resname + 'KeepDesiredSize', 'equals', 'k', resname + 'DesiredSize'),
                get_condition(resname + 'CapacityDesiredSize', 'equals', 'CapacityDesired', resname + 'DesiredSize'),
            ])
    
            # resources
            OutKey = ['DesiredSize']
   
            AS_ScheduledAction(resname, OutKey)
            r_ScheduledAction = ASScheduledAction(resname)
            r_ScheduledAction.setup()
    
            add_obj(r_ScheduledAction)
        

class AS_ScheduledActionsECS(object):
    def __init__(self, key):
        ScheduledActions = []
        for n, v in getattr(cfg, key).items():
            resname = '%s%s' % (key, n)
            # conditions
            c_disable = {resname + 'Disable': Or(
                Not(Condition(resname)),
                And(
                    Condition(resname + 'KeepMaxSize'),
                    Condition(resname + 'KeepMinSize'),
                )
            )}
    
            add_obj(c_disable)

            # resources
            AS_ScheduledAction(resname, [])
            ScheduledAction = APPASScheduledAction(resname)
            ScheduledAction.setup()

            ScheduledActions.append(
                If(
                    resname + 'Disable',
                    Ref('AWS::NoValue'),
                    ScheduledAction,
                )
            )

        self.ScheduledActions = ScheduledActions


class AS_ScalingPoliciesStepEC2(object):
    def __init__(self, key):
        R_Down = ASScalingPolicyStepDown('ScalingPolicyDown')
        R_Down.setup()

        R_Up = ASScalingPolicyStepUp('ScalingPolicyUp')
        R_Up.setup()

        add_obj([
            R_Down,
            R_Up,
        ])


class AS_ScalingPoliciesStepECS(object):
    def __init__(self, key):
        R_Down = APPASScalingPolicyStepDown('ScalingPolicyDown')
        R_Down.setup()

        R_Up = APPASScalingPolicyStepUp('ScalingPolicyUp')
        R_Up.setup()

        add_obj([
            R_Down,
            R_Up,
        ])


class AS_ScalingPoliciesTracking(object):
    def __init__(self, key):
        ScalingPolicyTrackings_Out_String = []
        ScalingPolicyTrackings_Out_Map = {}
        for n, v in getattr(cfg, key).items():
            if not ('Enabled' in v and v['Enabled'] is True):
                continue
            resname = '%s%s' % (key, n)
            # Autoscaling
            if 'TargetTrackingConfiguration' in v:
                TargetTrackingConfigurationName = 'TargetTrackingConfiguration'
                p_type = 'autoscaling'
            # Application Autoscaling
            elif 'TargetTrackingScalingPolicyConfiguration' in v:
                TargetTrackingConfigurationName = 'TargetTrackingScalingPolicyConfiguration'
                p_type = 'application_autoscaling'
            basename = resname + TargetTrackingConfigurationName

            # parameters
            p_Value = Parameter(basename + 'TargetValue')
            p_Value.Description = 'Tracking %s Value - 0 to disable - empty for default based on env/role' % n
            
            p_Statistic = Parameter(basename + 'CustomizedMetricSpecificationStatistic')
            p_Statistic.Description = 'Tracking %s Statistic - 0 to disable - empty for default based on env/role' % n

            add_obj([
                p_Value,
                p_Statistic,
            ])

            #  conditions
            add_obj(get_condition(resname, 'not_equals', '0', basename + 'TargetValue'))

            # outputs
            if v['Type'] == 'Cpu' or (v['Type'] == 'Custom' and
                v[TargetTrackingConfigurationName]['CustomizedMetricSpecification']['MetricName'] == 'CPUUtilization'
            ):
                ScalingPolicyTrackings_Out_String.append('Cpu${Statistic}:${Cpu}')
                ScalingPolicyTrackings_Out_Map.update({
                    'Statistic': get_endvalue(basename + 'CustomizedMetricSpecificationStatistic') if v['Type'] == 'Custom'  else '',
                    'Cpu': get_endvalue(basename + 'TargetValue'),
                })

            # resources
            r_Tracking = asg.ScalingPolicy(resname) if p_type == 'autoscaling' else aas.ScalingPolicy(resname)
            auto_get_props(r_Tracking, v, recurse=True)
            r_Tracking.Condition = resname

            add_obj(r_Tracking)

        # Outputs
        O_Policy = Output(key)
        O_Policy.Value = Sub(','.join(ScalingPolicyTrackings_Out_String), **ScalingPolicyTrackings_Out_Map)

        add_obj([
            O_Policy,
        ])


class AS_LaunchConfiguration(object):
    def __init__(self):
        InitConfigSets = ASInitConfigSets()
        InitConfigSets.setup()

        CfmInitArgs = {}
        IBoxEnvApp = []
        Tags = []
        UserDataApp = []

        for n in cfg.Apps:
            name = 'Apps%s' % n  # Ex. Apps1
            envname = 'EnvApp%sVersion' % n  # Ex EnvApp1Version
            reponame = name + 'RepoName'  # Ex Apps1RepoName

            UserDataApp.extend(['#${%s}\n' % envname])

            # parameters
            add_obj([
                Parameter(
                    envname,
                    Description='Application %s version' % n,
                    AllowedPattern='^[a-zA-Z0-9-_.]*$',
                ),
                Parameter(
                    reponame,
                    Description='App %s Repo Name - empty for default based on env/role' % n,
                    AllowedPattern='^[a-zA-Z0-9-_.]*$',
                )
            ])

            # conditions
            add_obj({
                name: And(
                    Not(Equals(Ref(envname), '')),
                    Not(get_condition('', 'equals', 'None', reponame))
                )
            })

            InitConfigApps = ASInitConfigApps(name)
            CfmInitArgs[name] = InitConfigApps

            InitConfigAppsBuilAmi = ASInitConfigAppsBuildAmi(name)
            # AUTOSPOT - Let cfn-init always prepare instances on boot
            #CfmInitArgs[name + 'BuildAmi'] = InitConfigAppsBuilAmi
            CfmInitArgs[name] = InitConfigAppsBuilAmi

            IBoxEnvApp.extend([
                'export EnvApp%sVersion=' % n, Ref(envname), "\n",
                'export EnvRepo%sName=' % n, get_endvalue(reponame), "\n",
            ])

            InitConfigSetsApp = If(name, name, Ref('AWS::NoValue'))
            InitConfigSetsAppBuilAmi = If(name, name + 'BuildAmi', Ref('AWS::NoValue'))
            IndexSERVICES = InitConfigSets.data['default'].index('SERVICES')
            InitConfigSets.data['default'].insert(IndexSERVICES, InitConfigSetsApp)
            # AUTOSPOT - Let cfn-init always prepare instances on boot
            #InitConfigSets.data['buildamifull'].append(InitConfigSetsAppBuilAmi)
            InitConfigSets.data['buildamifull'].append(InitConfigSetsApp)

            Tags.append(asg.Tag(envname, Ref(envname), True))

            # resources
            # FOR MULTIAPP CODEDEPLOY
            if len(cfg.Apps) > 1:
                r_DeploymentGroup = CDDeploymentGroup('DeploymentGroup' + name)
                r_DeploymentGroup.setup(index=n)

                add_obj(r_DeploymentGroup)

            # outputs
            Output_app = Output(envname)
            Output_app.Value = Ref(envname)
            add_obj(Output_app)

            Output_repo = Output(reponame)
            Output_repo.Value = get_endvalue(reponame)
            add_obj(Output_repo)

        InitConfigSetup = ASInitConfigSetup()
        InitConfigSetup.ibox_env_app = IBoxEnvApp
        InitConfigSetup.setup()

        InitConfigCodeDeploy = ASInitConfigCodeDeploy()
        InitConfigCodeDeploy.setup()

        CfmInitArgs['SETUP'] = InitConfigSetup

        if cfg.Apps:
            CfmInitArgs['CODEDEPLOY'] = InitConfigCodeDeploy
            CD_DeploymentGroup()


        # LoadBalancerClassic External
        if cfg.LoadBalancerClassicExternal:
            InitConfigELBExternal = ASInitConfigELBClassicExternal()
            InitConfigELBExternal.setup()
            CfmInitArgs['ELBWAITER'] = InitConfigELBExternal

        # LoadBalancerClassic Internal
        if cfg.LoadBalancerClassicInternal:
            InitConfigELBInternal = ASInitConfigELBClassicInternal()
            InitConfigELBInternal.setup()
            CfmInitArgs['ELBWAITER'] = InitConfigELBInternal

        # LoadBalancerApplication External
        if cfg.LoadBalancerApplicationExternal:
            InitConfigELBExternal = ASInitConfigELBApplicationExternal()
            InitConfigELBExternal.setup()
            CfmInitArgs['ELBWAITER'] = InitConfigELBExternal

        # LoadBalancerApplication Internal
            InitConfigELBInternal = ASInitConfigELBApplicationInternal()
            InitConfigELBInternal.setup()
            CfmInitArgs['ELBWAITER'] = InitConfigELBInternal

        SecurityGroups = SG_SecurityGroupsEC2().SecurityGroups

        # Resources
        R_LaunchConfiguration = ASLaunchConfiguration('LaunchConfiguration')
        R_LaunchConfiguration.setup(UserDataApp=UserDataApp)
        R_LaunchConfiguration.SecurityGroups.extend(SecurityGroups)

        R_InstanceProfile = IAMInstanceProfile('InstanceProfile')
        R_InstanceProfile.setup()

        # Import role specific cfn definition
        if 'cfn_' + cfg.classenvrole in globals():  # Ex cfn_client_portal
            CfnRole = globals()['cfn_' + cfg.classenvrole]()
            CfmInitArgs.update(CfnRole)

        R_LaunchConfiguration.Metadata = cfm.Metadata(
            cfm.Init(
                InitConfigSets,
                **CfmInitArgs
            ),
            cfm.Authentication({
                'CfnS3Auth': cfm.AuthenticationBlock(
                    type='S3',
                    buckets=[
                        Sub(cfg.BucketAppRepository),
                        Sub(cfg.BucketAppData)
                    ],
                    roleName=Ref('RoleInstance')
                )
            })
        )

        R_LaunchConfigurationSpot = ASLaunchConfiguration('LaunchConfigurationSpot')
        R_LaunchConfigurationSpot.setup(UserDataApp=UserDataApp, spot=True)
        R_LaunchConfigurationSpot.SecurityGroups = R_LaunchConfiguration.SecurityGroups
        R_LaunchConfigurationSpot.SpotPrice = get_endvalue('SpotPrice')

        add_obj([
            R_LaunchConfiguration,
            R_InstanceProfile,
        ]) 
        if cfg.SpotASG:
            add_obj(R_LaunchConfigurationSpot)

        self.LaunchConfiguration = R_LaunchConfiguration
        self.Tags = Tags


class AS_AutoscalingEC2(object):
    def __init__(self, key):
        LoadBalancers = []
        for n in cfg.LoadBalancerClassic:
            LoadBalancers.append(Ref('LoadBalancerClassic' + n))

        TargetGroups = []
        for n in cfg.LoadBalancerApplication:
            TargetGroups.append(Ref('TargetGroup' + n))

        # Resources
        AS_ScheduledActionsEC2('ScheduledAction')
        #AS_ScalingPoliciesEC2()

        LaunchConfiguration = AS_LaunchConfiguration()
        Tags = LaunchConfiguration.Tags

        R_ASG = ASAutoScalingGroup('AutoScalingGroup')
        R_ASG.setup()
        R_ASG.LoadBalancerNames = LoadBalancers
        R_ASG.TargetGroupARNs = TargetGroups
        R_ASG.Tags.extend(Tags)
        R_ASG.Tags.extend([
            If(
                'SpotAuto',
                asg.Tag(('spot-enabled'), 'true', True),
                Ref('AWS::NoValue')
            ),
            If(
                'SpotAutoMinOnDemandNumber',
                asg.Tag(('autospotting_min_on_demand_number'), get_endvalue('SpotAutoMinOnDemandNumber'), True),
                Ref('AWS::NoValue')
            ),
            If(
                'SpotAutoAllowedInstances',
                asg.Tag(('autospotting_allowed_instance_types'), get_endvalue('SpotAutoAllowedInstances'), True),
                Ref('AWS::NoValue')
            )
        ])

        R_ASGSpot = ASAutoScalingGroup('AutoScalingGroupSpot')
        R_ASGSpot.setup(spot=True)
        R_ASGSpot.LoadBalancerNames = LoadBalancers
        R_ASGSpot.TargetGroupARNs = TargetGroups
        R_ASGSpot.Tags.extend(Tags)

        # Notifications currently are not associeted to "any actions" - now using CW events - this way works with autospotting too
        try: cfg.NotificationConfiguration
        except:
            pass
        else:
            NotificationConfiguration = ASNotificationConfiguration()
            NotificationConfiguration.setup()
            R_ASG.NotificationConfigurations = [NotificationConfiguration]
            R_ASGSpot.NotificationConfigurations = [NotificationConfiguration]

        add_obj([
            R_ASG,
        ])

        if cfg.SpotASG:
            add_obj(R_ASGSpot)

        self.LaunchConfiguration = LaunchConfiguration


class AS_AutoscalingECS(object):
    def __init__(self, key):
        # Resources
        R_ScalableTarget = APPASScalableTarget('ScalableTarget')
        R_ScalableTarget.setup()
        R_ScalableTarget.ScheduledActions = AS_ScheduledActionsECS('ScheduledAction').ScheduledActions

        add_obj([
            R_ScalableTarget
        ])


class AS_LifecycleHook(object):
    def __init__(self,key):
        for n, v in getattr(cfg, key).items():
            resname = '%s%s' % (key, n)

            # resources
            r_Hook = ASLifecycleHook(resname, name=n, key=v)
            r_Hook.AutoScalingGroupName = Ref('AutoScalingGroup')

            r_HookSpot = ASLifecycleHook(resname, name=n, key=v)
            r_HookSpot.title = resname + 'Spot'
            r_HookSpot.Condition = 'SpotASG'
            r_HookSpot.AutoScalingGroupName = Ref('AutoScalingGroupSpot')
                
            add_obj([
                r_Hook,
            ])

            if cfg.SpotASG:
                add_obj(r_HookSpot)



import yaml
import yaml.constructor
import sys
import os
import copy
import json
import logging
from collections import OrderedDict
from collections.abc import Mapping
from pprint import pprint, pformat

from . import cfg


def show_odict(odict):
    print(json.dumps(odict, indent=4))


# We also need to load mappings in order
# Based on https://gist.github.com/844388
def construct_ordereddict(loader, node):
    data = OrderedDict()
    yield data
    value = construct_mapping(loader, node)
    data.update(value)


def construct_mapping(self, node, deep=False):
    if isinstance(node, yaml.MappingNode):
        self.flatten_mapping(node)
    else:
        raise yaml.constructor.ConstructorError(
            None, None, 'expected a mapping node, but found %s' % node.id, node.start_mark)

    mapping = OrderedDict()
    for key_node, value_node in node.value:
        key = self.construct_object(key_node, deep=deep)
        try:
            hash(key)
        except TypeError as exc:
            raise yaml.constructor.ConstructorError(
                'while constructing a mapping', node.start_mark,
                'found unacceptable key (%s)' % exc, key_node.start_mark)
        value = self.construct_object(value_node, deep=deep)
        mapping[key] = value
    return mapping


class Loader(yaml.Loader):
    def __init__(self, stream):
        # This way for include relative to file with include statement
        self._root_current = os.path.split(stream.name)[0]
        # This way for include BASE relative on BASE dir
        self._root_base = os.path.join(CFG_FILE_INT, 'BASE')
        # This way for include relative on BASE EXT dir
        self._root_brand_base = os.path.join(CFG_FILE_EXT, 'BASE')
        # This way for include relative on brand EXT dir
        self._root_brand_ext = os.path.join(CFG_FILE_EXT, brand)
        self.stream = stream
        super(Loader, self).__init__(stream)
        Loader.add_constructor('!include', Loader.include)
        Loader.add_constructor('!import',  Loader.include)
        Loader.add_constructor('!exclude',  Loader.exclude)


    def exclude(self, node):
        global exclude_list
        if   isinstance(node, yaml.ScalarNode):
            exclude_list.append(node)
        elif isinstance(node, yaml.SequenceNode):
            for filename in self.construct_sequence(node):
                exclude_list.append(filename)


    def include(self, node):
        if isinstance(node, yaml.ScalarNode) and node not in exclude_list:
            return self.extractFile(self.construct_scalar(node), self._root_current)

        elif isinstance(node, yaml.SequenceNode):
            result = []
            for filename in self.construct_sequence(node):
                if filename in exclude_list:
                    continue
                # if including Ext cfg try to include BASE file from CFG_FILE_INT too
                if CFG_FILE_EXT in self.stream.name:
                    result.append(self.extractFile(filename, self._root_base))
                result.append(self.extractFile(filename, self._root_current))
                # try to include BASE ext too
                result.append(self.extractFile(filename, self._root_brand_base))
                # try to include brand ext too
                result.append(self.extractFile(filename, self._root_brand_ext))
                # result += self.extractFile(filename)
            return result

        elif isinstance(node, yaml.MappingNode):
            result = {}
            for k,v in self.construct_mapping(node).items():
                if k in exclude_list:
                    continue
                result[k] = self.extractFile(v, self._root_current)
            return result

        else:
            print('Error:: unrecognised node type in !include statement')
            raise yaml.constructor.ConstructorError

    def extractFile(self, filename, root):
        filepath = os.path.join(root, filename)
        try:
            with open(filepath, 'r') as f:
                return yaml.load(f, Loader)
        except IOError:
            pass


def replace_not_allowed_char(s):
    key = str(s)
    for s, w in {
            '/': 'SLASH',
            '*': 'STAR',
            '-': 'HYPH',
            '?': 'QUEST',
            '.': 'DOT',
            '_': 'USCORE',
    }.items():
        key = key.replace(s, w)

    return int(key) if key.isdigit() else key


def gen_dict_extract(cfg, envs):
    global enforce_list
    if hasattr(cfg, 'items'):
        # This method allow to delete items from a dictionary while iterating over it
        for k in list(cfg):
            v = cfg[k]
            # for final values
            if isinstance(v, (str, int, list)) and not k.startswith('IBoxLoader'):
                if k in enforce_list:
                    continue
                elif k.endswith('_IBOXENFORCE'):
                    k = k.replace('_IBOXENFORCE', '')
                    enforce_list.append(k)
                yield {k: v}
            # for empty dict in common.yml
            if isinstance(v, dict) and len(v) == 0:
                yield {k: v}
            # for recursively descending in env/region role dict
            if k in envs and isinstance(v, (dict, list)):
                try:
                    # after descending in env main key (not the one nested under region) delete key
                    # this way when envs include both (env and region) i do not process it again
                    if k in list(RP_base.keys()):
                        del cfg[k]
                except:
                    pass
                for result in gen_dict_extract(v, envs):
                    yield result
            # for recursively descending in dict not in RP_base_keys (env/region/envrole/stacktype) 
            # (final key is the concatenation of traversed dict keys)
            if k not in RP_base_keys and isinstance(v, (dict)):
                for j, w in v.items():
                    for result in gen_dict_extract({k + j: w}, envs):
                        yield result
    if isinstance(cfg, list):
        for n in cfg:
            for result in gen_dict_extract(n, envs):
                yield result


def my_merge_dict(basedict, workdict):
    if isinstance(workdict, (str, list)):
        return workdict
    if len(basedict) > 0:
        sumdict = dict(list(basedict.items()) + list(workdict.items()))
    else:
        return dict(list(workdict.items()))

    for k in sumdict.keys():
        try:
            is_dict = isinstance(workdict[k], dict)
            is_map = isinstance(basedict[k], Mapping)
        except:
            is_dict = False
            is_map = False

        # Trick to be able to add an element to a previuosly created/defined list.
        # I can centralize some general iampolicy in ecs-cluster and attach them to RoleInstance,
        # adding element to ManagedPolicyArns, preserving the base ones
        try:
            ibox_add_to_list = (workdict[k][0] == 'IBOXADDTOLIST')
        except:
            ibox_add_to_list = False

        if is_dict and is_map:
            my_merge_dict(basedict[k], workdict[k])
        elif ibox_add_to_list:
            del workdict[k][0]
            basedict[k] += workdict[k]
        elif k in workdict:
            basedict[k] = workdict[k]
        else:
            pass

    return basedict


def get_RP_for_envs(value):
    RP = OrderedDict()
    
    try:
        is_dict = isinstance(value[0], (OrderedDict, dict))
    except:
        is_dict = False

    if hasattr(value, 'items'):
        for d, v in value.items():
            RP[d] = get_RP_for_envs(v)
    elif is_dict:
        for d, v in enumerate(value):
            for i, j in v.items():
                # CF Mapping allow for index only alfanumeric char,
                # this way i can specify more "clear" name for index in CloudFormation behaviours
                key = replace_not_allowed_char(i)
                # RP[key] already exist as a dict, try merging
                if key in RP and isinstance(RP[key], dict):
                    RP[key] = my_merge_dict(RP[key], get_RP_for_envs(j))
                else:
                    RP[key] = get_RP_for_envs(j)
    elif isinstance(value, list):
        RP = list(value)
    else:
        RP = value

    return RP


def read_yaml(file_type, brand, base_dir):
    cfg_file = os.path.join(base_dir, brand, file_type + '.yml')

    yaml.Loader.add_constructor('tag:yaml.org,2002:map', construct_ordereddict)
    try:
        with open(cfg_file, 'r') as ymlfile:
            cfg = yaml.load(ymlfile, Loader=Loader)

            return cfg
    except IOError:
        return {}


def parse_cfg(cfg, envs=[]):
    odict = OrderedDict([])

    for value in gen_dict_extract(cfg, envs):
        for k, v in value.items():
            try:
                if isinstance(v[0], dict):
                    odict[k] = odict[k] + v
                else:
                    raise
            except:
                odict[k] = v

    return odict


def merge_cfg(cfgs, cfg_key, list_base=None):
    RP_list = copy.deepcopy(list_base) if list_base else []

    for cfg, v in cfgs.items():
        for c in v:
            keys = ['IBoxLoader', 'IBoxLoaderAfter'] + cfg_key[cfg]
            parsed_cfg = parse_cfg(c, keys)
            RP_list.append(parsed_cfg)

    return RP_list


def prepend_base_cfgs(cfg_cmm):
    for cfg_key, cfg_value in cfg.BASE_CFGS.items():
        key_names = []
        for c in cfg_cmm:
            if cfg_key in c:
                for k in c[cfg_key]:
                    key_names.extend(list(k.keys()))

        if key_names:
            keys = []
            for k in key_names:
                keys.append({
                    k: cfg_value
                })

            cfg_cmm.insert(0, {cfg_key: keys})


def get_RP(cfgs):
    RP = copy.deepcopy(RP_base)

    cfg_key_cmm = {
        'common': ['global'],
        'type': [stacktype, 'global'],
        'role': [envrole, 'global'],
    }

    RP_list = []

    cfg_merge_cmm = merge_cfg(cfgs, cfg_key_cmm)

    # Prepend base config from awsibox/cfg.py BASE_CFGS
    prepend_base_cfgs(cfg_merge_cmm)

    RP['cmm']['cmm'] = get_RP_for_envs(cfg_merge_cmm)

    for env, rvalue in RP.items():
        if env == 'cmm':
            continue

        cfg_key_env = {
            'common': [env],
            'type': [env],
            'role': [env],
        }

        cfg_merge_env = merge_cfg(cfgs, cfg_key_env)

        for region in rvalue.keys():

            cfg_key_region = {
                'common': [region],
                'type': [region],
                'role': [region],
            }

            cfg_key_env_region = {
                'common': [env, region],
                'type': [env, region],
                'role': [env, region],
            }

            cfg_merge_region = merge_cfg(cfgs, cfg_key_region, cfg_merge_env)
            cfg_merge_env_region = merge_cfg(cfgs, cfg_key_env_region, cfg_merge_region)

            RP[env][region] = get_RP_for_envs(cfg_merge_env_region)

    return(RP)


def get_RP_base_keys():
    RP_base_keys = ['global']
    for n, v in RP_base.items():
        RP_base_keys.append(n)
        for m, w in v.items():
            RP_base_keys.append(m)

    return RP_base_keys + [stacktype, envrole]


def get_stack_type(cfgs):
    for c in cfgs:
        try:
            if isinstance(c[envrole], dict) and 'StackType' in c[envrole]:
                return c[envrole]['StackType']
        except KeyError:
            pass


def set_cfg():
    for n, v in cfg.RP_cmm.items():
        setattr(cfg, n, v)
    
    # set generic attribute based on condition:
    
    # LoadBalancerClassic
    try: cfg.LoadBalancerClassic
    except:
        cfg.LoadBalancerClassic = []
    
    # LoadBalancerApplication
    try: cfg.LoadBalancerApplication
    except:
        cfg.LoadBalancerApplication = []
    
    # LoadBalancer
    cfg.LoadBalancer = None
    for n in ['LoadBalancerClassic', 'LoadBalancerApplication']:
        try: getattr(cfg, n)
        except:
            pass
        else:
            cfg.LoadBalancer = True
    
    # LoadBalancerClassicExternal LoadBalancerClassicInternal
    cfg.LoadBalancerClassicExternal = None
    cfg.LoadBalancerClassicInternal = None
    try: cfg.LoadBalancerClassic
    except:
        pass
    else:
        if 'External' in cfg.LoadBalancerClassic:
            cfg.LoadBalancerClassicExternal = True
        if 'Internal' in cfg.LoadBalancerClassic:
            cfg.LoadBalancerClassicInternal = True
    
    # LoadBalancerApplicationExternal LoadBalancerApplicationInternal
    cfg.LoadBalancerApplicationExternal = None
    cfg.LoadBalancerApplicationInternal = None
    try: cfg.LoadBalancerApplication
    except:
        pass
    else:
        if 'External' in cfg.LoadBalancerApplication:
            cfg.LoadBalancerApplicationExternal = True
        if 'Internal' in cfg.LoadBalancerApplication:
            cfg.LoadBalancerApplicationInternal = True
    
    # SpotASG
    try: cfg.SpotASG
    except:
        cfg.SpotASG = None
    
    # RecordSet
    cfg.RecordSetExternal = None
    cfg.RecordSetInternal = None
    if 'External' in cfg.RecordSet:
        cfg.RecordSetExternal = True
    if 'Internal' in cfg.RecordSet:
        cfg.RecordSetInternal = True
    
    cfg.stacktype = stacktype


def build_RP():
    global envrole
    global stacktype
    global RP_base
    global RP_base_keys
    global brand
    global CFG_FILE_INT
    global CFG_FILE_EXT
    global enforce_list
    global exclude_list

    enforce_list = []
    exclude_list = []

    CFG_FILE_INT = '%s/cfg' % os.path.dirname(os.path.realpath(__file__))
    CFG_FILE_INT = os.path.normpath(CFG_FILE_INT)
    
    CFG_FILE_EXT = '%s/cfg' % os.getcwd()
    CFG_FILE_EXT = os.path.normpath(CFG_FILE_EXT)
    
    envrole = cfg.envrole
    brand = cfg.brand
    
    RP_base = OrderedDict([
        ('cmm', {
            'cmm': {},
        }),
    ])
    
    # dynamically create RP_BASE OrderedDict from ENV_BASE and cfg.regions
    for n in cfg.ENV_BASE:
        RP_base[n] = {}
        for m in cfg.regions:
            RP_base[n][m] = {}

    cfg.RP_base = RP_base
    
    cfg_role = [
        read_yaml(envrole, 'BASE', CFG_FILE_INT),
        read_yaml(envrole, 'BASE', CFG_FILE_EXT),
        read_yaml(envrole, brand, CFG_FILE_EXT),
    ]
    
    stacktype = get_stack_type(cfg_role)
    
    if not stacktype:
        logging.error('StackType key not found for Role %s in paths: %s, %s', envrole, CFG_FILE_EXT, CFG_FILE_INT)
        exit(1)
    
    cfgs = OrderedDict([
        ('common', [
            read_yaml('common', 'BASE', CFG_FILE_INT),
            read_yaml('common', 'BASE', CFG_FILE_EXT),
            read_yaml('common', brand, CFG_FILE_EXT),
        ]),
        ('type', [
            read_yaml(stacktype, 'BASE', CFG_FILE_INT),
            read_yaml(stacktype, 'BASE', CFG_FILE_EXT),
            read_yaml(stacktype, brand, CFG_FILE_EXT),
        ]),
        ('role', cfg_role),
    ])
    
    RP_base_keys = get_RP_base_keys()
    
    RP = get_RP(cfgs)
    
    # print(RP['dev']['eu-west-1']['CloudFrontCacheBehaviors'][2]['QueryStringCacheKeys'])
    if cfg.debug:
        print('##########RP#########START#####')
        # TO DEBUG - NICELY PRINT NESTED ORDEREDDICT
        show_odict(RP)
        print('##########RP#########END#######')
        print('##########ENFORCED######START#####')
        pprint(enforce_list)
        print('##########ENFORCED######END#######')
        print('##########EXCLUDE#######START#####')
        pprint(exclude_list)
        print('##########EXCLUDE#######END#######')


    try:
        stacktype = RP['cmm']['cmm']['StackType']
    except KeyError:
        print('StackType key not found!')
        exit(1)

    cfg.RP = RP
    cfg.RP_cmm = RP['cmm']['cmm']

    set_cfg()
